// $Id: Hurd288Engine.cpp,v 1.1.1.1 2000/04/03 21:03:44 wilko Exp $
// -*- C++ -*-
//
// -----------------------------------------------------------------------
//                           HEP Random
//                     --- Hurd288Engine ---
//                    class implementation file
// -----------------------------------------------------------------------
// An interconnected shift register based on the paper presented by Hurd in
// IEEE Transactions on Computers c23, 2 Feb 1974.
// =======================================================================
// Ken Smith      - Initial draft started:                    23rd Jul 1998
//                - Added conversion operators:                6th Aug 1998
// J. Marraffino  - Added some explicit casts to deal with
//                  machines where sizeof(int) != sizeof(long)  22 Aug 1998
// =======================================================================

#include "CLHEP/Random/Random.h"
#include "CLHEP/Random/Hurd288Engine.h"
//#include "CLHEP/String/Strings.h"
#include <iomanip>

// Number of instances with automatic seed selection
HepInt Hurd288Engine::numEngines = 0;

// Maximum index into the seed table
HepInt Hurd288Engine::maxIndex = 215;

static inline unsigned int f288(unsigned int a, unsigned int b, unsigned int c)
{
  return ( ((b<<2) & 0x7ffc) | ((a<<2) & ~0x7ffc) | (a>>30) ) ^ 
         ( (c<<1)|(c>>31) );
}

Hurd288Engine::Hurd288Engine() 
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.) )
{
  HepInt cycle    = abs(HepInt(numEngines/maxIndex));
  HepInt curIndex = abs(HepInt(numEngines%maxIndex));
  long mask = ((cycle & 0x007fffff) << 8);
  long seedlist[2];
  HepRandom::getTheTableSeeds( seedlist, curIndex );
  seedlist[0] ^= mask;
  seedlist[1] = 0;
  setSeeds(seedlist, 0);
  ++numEngines;
  for( HepInt i=0; i < 100; ++i ) flat();       // warm up just a bit
}

Hurd288Engine::Hurd288Engine( std::istream& is )
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.) )
{
    is >> *this;
}

Hurd288Engine::Hurd288Engine( long seed )
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.) )
{
  long seedlist[2]={seed,0};
  setSeeds(seedlist, 0);
  for( HepInt i=0; i < 100; ++i ) flat();       // warm up just a bit
}

Hurd288Engine::Hurd288Engine( HepInt rowIndex, HepInt colIndex )
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.) )
{
  HepInt cycle = abs(HepInt(rowIndex/maxIndex));
  HepInt   row = abs(HepInt(rowIndex%maxIndex));
  HepInt   col = colIndex & 0x1;
  long mask = (( cycle & 0x000007ff ) << 20 );
  long seedlist[2];
  HepRandom::getTheTableSeeds( seedlist, row );
  seedlist[0] = (seedlist[col])^mask;
  seedlist[1]= 0;
  setSeeds(seedlist, 0);
  for( HepInt i=0; i < 100; ++i ) flat();       // warm up just a bit
}

Hurd288Engine::~Hurd288Engine() { }

Hurd288Engine::Hurd288Engine( const Hurd288Engine & p ) 
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.) )
{
  *this = p;
}

Hurd288Engine & Hurd288Engine::operator=( const Hurd288Engine & p ) {
  if( this != &p ) {
    wordIndex = p.wordIndex;
    for( HepInt i=0; i < 9; ++i ) {
      words[i] = p.words[i];
    }
  }
  return *this;
}

HepDouble Hurd288Engine::flat() {

  if( wordIndex <= 0 ) {
     register unsigned int W0, W1, W2, W3, W4, W5, W6, W7, W8;

     W0 = words[0];
     W1 = words[1];
     W2 = words[2];
     W3 = words[3];
     W4 = words[4];
     W5 = words[5];
     W6 = words[6];
     W7 = words[7];
     W8 = words[8];
     W1 ^= W0; W0 = f288(W2, W3, W0);
     W2 ^= W1; W1 = f288(W3, W4, W1);
     W3 ^= W2; W2 = f288(W4, W5, W2);
     W4 ^= W3; W3 = f288(W5, W6, W3);
     W5 ^= W4; W4 = f288(W6, W7, W4);
     W6 ^= W5; W5 = f288(W7, W8, W5);
     W7 ^= W6; W6 = f288(W8, W0, W6);
     W8 ^= W7; W7 = f288(W0, W1, W7);
     W0 ^= W8; W8 = f288(W1, W2, W8);
     words[0] = W0 & 0xffffffff;
     words[1] = W1 & 0xffffffff;
     words[2] = W2 & 0xffffffff;
     words[3] = W3 & 0xffffffff;
     words[4] = W4 & 0xffffffff;
     words[5] = W5 & 0xffffffff;
     words[6] = W6 & 0xffffffff;
     words[7] = W7 & 0xffffffff;
     words[8] = W8 & 0xffffffff;
     wordIndex = 9;
  }

  return  words[       --wordIndex] * mantissa_bit_32 + // most significant part
     (words[(wordIndex+2)&0x7]>>11) * mantissa_bit_53 + // fill in rest of bits
                    mantissa_bit_54 - mantissa_bit_106; // make sure non-zero
}

void Hurd288Engine::flatArray( const HepInt size, HepDouble* vect ) {
    for (HepInt i = 0; i < size; ++i) {
        vect[i] = flat();
    }
}

void Hurd288Engine::setSeed( long seed, HepInt ) {
  words[0] = (unsigned int)seed;
  for (wordIndex = 1; wordIndex < 9; ++wordIndex) {
    words[wordIndex] = 69607 * words[wordIndex-1] + 54329;
  }
}

void Hurd288Engine::setSeeds( const long* seeds, HepInt ) {
  setSeed( *seeds ? *seeds : 32767, 0 );
  theSeeds = seeds;
}
     
void Hurd288Engine::saveStatus( const char filename[] ) const {
  std::ofstream outFile(filename, std::ios::out);
  if( !outFile.bad() ) {
    outFile << std::setprecision(20) << theSeed << " ";
    outFile << wordIndex << " ";
    for( HepInt i = 0; i < 9; ++i ) {
      outFile << words[i] << " ";
    }
  }
}

void Hurd288Engine::restoreStatus( const char filename[] ) {
  std::ifstream inFile(filename, std::ios::in);
  if( !inFile.bad() ) {
    inFile >> theSeed;
    inFile >> wordIndex;
    for( HepInt i = 0; i < 9; ++i ) {
        inFile >> words[i];
    }
  }
}

void Hurd288Engine::showStatus() const {
  std::cout << std::setprecision(20) << std::endl;
  std::cout << "----------- Hurd2 engine status ----------" << std::endl;
  std::cout << "Initial seed  = " << theSeed   << std::endl;
  std::cout << "Current index = " << wordIndex << std::endl;
  std::cout << "Current words = " << std::endl;
  for( HepInt i = 0; i < 9 ; ++i ) {
    std::cout << "    " << words[i] << std::endl;
  }
  std::cout << "-------------------------------------------" << std::endl;
}

Hurd288Engine::operator HepFloat() {
  if( wordIndex <= 0 ) {
     register unsigned int W0, W1, W2, W3, W4, W5, W6, W7, W8;

     W0 = words[0];
     W1 = words[1];
     W2 = words[2];
     W3 = words[3];
     W4 = words[4];
     W5 = words[5];
     W6 = words[6];
     W7 = words[7];
     W8 = words[8];
     W1 ^= W0; W0 = f288(W2, W3, W0);
     W2 ^= W1; W1 = f288(W3, W4, W1);
     W3 ^= W2; W2 = f288(W4, W5, W2);
     W4 ^= W3; W3 = f288(W5, W6, W3);
     W5 ^= W4; W4 = f288(W6, W7, W4);
     W6 ^= W5; W5 = f288(W7, W8, W5);
     W7 ^= W6; W6 = f288(W8, W0, W6);
     W8 ^= W7; W7 = f288(W0, W1, W7);
     W0 ^= W8; W8 = f288(W1, W2, W8);
     words[0] = W0 & 0xffffffff;
     words[1] = W1 & 0xffffffff;
     words[2] = W2 & 0xffffffff;
     words[3] = W3 & 0xffffffff;
     words[4] = W4 & 0xffffffff;
     words[5] = W5 & 0xffffffff;
     words[6] = W6 & 0xffffffff;
     words[7] = W7 & 0xffffffff;
     words[8] = W8 & 0xffffffff;
     wordIndex = 9;
  }
  return words[--wordIndex ] * mantissa_bit_32;
}

Hurd288Engine::operator unsigned int() {
  if( wordIndex <= 0 ) {
     register unsigned int W0, W1, W2, W3, W4, W5, W6, W7, W8;

     W0 = words[0];
     W1 = words[1];
     W2 = words[2];
     W3 = words[3];
     W4 = words[4];
     W5 = words[5];
     W6 = words[6];
     W7 = words[7];
     W8 = words[8];
     W1 ^= W0; W0 = f288(W2, W3, W0);
     W2 ^= W1; W1 = f288(W3, W4, W1);
     W3 ^= W2; W2 = f288(W4, W5, W2);
     W4 ^= W3; W3 = f288(W5, W6, W3);
     W5 ^= W4; W4 = f288(W6, W7, W4);
     W6 ^= W5; W5 = f288(W7, W8, W5);
     W7 ^= W6; W6 = f288(W8, W0, W6);
     W8 ^= W7; W7 = f288(W0, W1, W7);
     W0 ^= W8; W8 = f288(W1, W2, W8);
     words[0] = W0 & 0xffffffff;
     words[1] = W1 & 0xffffffff;
     words[2] = W2 & 0xffffffff;
     words[3] = W3 & 0xffffffff;
     words[4] = W4 & 0xffffffff;
     words[5] = W5 & 0xffffffff;
     words[6] = W6 & 0xffffffff;
     words[7] = W7 & 0xffffffff;
     words[8] = W8 & 0xffffffff;
     wordIndex = 9;
  }
  return words[--wordIndex];
}
/*
std::ostream& operator<< (std::ostream& os, const Hurd288Engine& e) {
  HepString beginMarker = "Hurd288Engine-begin";
  HepString endMarker   = "Hurd288Engine-end";

  os << " " << beginMarker << " ";

  os << std::setprecision(20)  << e.theSeed  << " ";
  os << e.wordIndex << " ";
  for (HepInt i = 0; i < 9; ++i) {
    os << e.words[i]  << " ";
  }
  os << " " << endMarker   << " ";
  return os;
}
*/
/*
std::istream& operator>> (std::istream& is, Hurd288Engine& e) {
  HepString beginMarker = " ";
  HepString endMarker   = " ";

  is >> std::ws >> beginMarker;
  if (beginMarker != "Hurd288Engine-begin") {
    is.clear(std::ios::badbit | is.rdstate());
    std::cerr << "\nInput mispositioned or"
	      << "\nHurd288Engine state description missing or"
	      << "\nwrong engine type found." << std::endl;
    return is;
  }
  is >> e.theSeed >> e.wordIndex;
  for (HepInt i = 0; i < 9; ++i) {
    is >> e.words[i];
  }
  is >> std::ws >> endMarker;
  if (endMarker != "Hurd288Engine-end") {
    is.clear(std::ios::badbit | is.rdstate());
    std::cerr << "\nHurd288Engine state description incomplete."
	      << "\nInput stream is probably mispositioned now." << std::endl;
    return is;
  }
  return is;
}
*/
