// $Id: JamesRandom.cpp,v 1.1.1.1 2000/04/03 21:03:44 wilko Exp $
// -*- C++ -*-
//
// -----------------------------------------------------------------------
//                             HEP Random
//                       --- HepJamesRandom ---
//                      class implementation file
// -----------------------------------------------------------------------
// This file is part of Geant4 (simulation toolkit for HEP).
//
// This algorithm implements the original universal random number generator
// as proposed by Marsaglia & Zaman in report FSU-SCRI-87-50 and coded
// in FORTRAN77 by Fred James as the RANMAR generator, part of the MATHLIB
// HEP library.

// =======================================================================
// Gabriele Cosmo - Created: 5th September 1995
//                - Fixed a bug in setSeed(): 26th February 1996
//                - Minor corrections: 31st October 1996
//                - Added methods for engine status: 19th November 1996
//                - Fixed bug in setSeeds(): 15th September 1997
// J.Marraffino   - Added stream operators and related constructor.
//                  Added automatic seed selection from seed table and
//                  engine counter: 16th Feb 1998
// Ken Smith      - Added conversion operators:  6th Aug 1998
// =======================================================================

#include "CLHEP/Random/Random.h"
#include "CLHEP/Random/JamesRandom.h"
//#include "CLHEP/String/Strings.h"

#include <iomanip>
// Number of instances with automatic seed selection
HepInt HepJamesRandom::numEngines = 0;

// Maximum index into the seed table
HepInt HepJamesRandom::maxIndex = 215;

HepJamesRandom::HepJamesRandom(long seed)
{
  setSeed(seed,0);
  setSeeds(&theSeed,0);
}

HepJamesRandom::HepJamesRandom()     	// 15 Feb. 1998  JMM
{
  long seeds[2];
  long seed;
  HepInt cycle,curIndex;

  cycle = abs(HepInt(numEngines/maxIndex));
  curIndex = abs(HepInt(numEngines%maxIndex));
  numEngines += 1;
  long mask = ((cycle & 0x007fffff) << 8);
  HepRandom::getTheTableSeeds( seeds, curIndex );
  seed = seeds[0]^mask;
  setSeed(seed,0);
  setSeeds(&theSeed,0);
}

HepJamesRandom::HepJamesRandom(HepInt rowIndex, HepInt colIndex) // 15 Feb. 1998  JMM
{
  long seed;
   long seeds[2];

  HepInt cycle = abs(HepInt(rowIndex/maxIndex));
  HepInt row = abs(HepInt(rowIndex%maxIndex));
  HepInt col = abs(HepInt(colIndex%2));
  long mask = ((cycle & 0x000007ff) << 20);
  HepRandom::getTheTableSeeds( seeds, row );
  seed = (seeds[col])^mask;
  setSeed(seed,0);
  setSeeds(&theSeed,0);
}

#if 0
HepJamesRandom::HepJamesRandom(std::istream& is)
{
  is >> *this;
}

#endif
 
HepJamesRandom::~HepJamesRandom() {}

HepJamesRandom::HepJamesRandom(const HepJamesRandom &p)
{
  HepInt ipos, jpos;
  if ((this != &p) && (&p)) {
    theSeed = p.getSeed();
    setSeeds(&theSeed,0);
    for (HepInt i=0; i<97; ++i)
      u[i] = p.u[i];
    c = p.c; cd = p.cd; cm = p.cm;
    jpos = HepInt(p.pj97-p.u);
    ipos = (64+jpos)%97;
    pi97 = &u[ipos]; pj97 = &u[jpos];
  }
}

HepJamesRandom & HepJamesRandom::operator = (const HepJamesRandom &p)
{
  HepInt ipos, jpos;
  if ((this != &p) && (&p)) {
    theSeed = p.getSeed();
    setSeeds(&theSeed,0);
    for (HepInt i=0; i<97; ++i)
      u[i] = p.u[i];
    c = p.c; cd = p.cd; cm = p.cm;
    jpos = HepInt(p.pj97-p.u);
    ipos = (64+jpos)%97;
    pi97 = &u[ipos]; pj97 = &u[jpos];
  }
  return *this;
}

void HepJamesRandom::saveStatus( const char filename[] ) const
{
   std::ofstream outFile( filename, std::ios::out ) ;

   if (!outFile.bad()) {
     HepInt pos = HepInt(pj97-u);
     outFile << theSeed << std::endl;
     for (HepInt i=0; i<97; ++i)
     outFile << std::setprecision(20) << u[i] << " ";
     outFile << std::endl;
     outFile << std::setprecision(20) << c << " ";
     outFile << std::setprecision(20) << cd << " ";
     outFile << std::setprecision(20) << cm << std::endl;
     outFile << pos << std::endl;
   }
}

void HepJamesRandom::restoreStatus( const char filename[] )
{
   HepInt ipos, jpos;
   std::ifstream inFile( filename, std::ios::in);

#ifdef _MSC_VER
   if( inFile.is_open() && !inFile.eof() ) {
#else
   if (!inFile.bad() && !inFile.eof()) {
#endif
     inFile >> theSeed;
     for (HepInt i=0; i<97; ++i)
       inFile >> u[i];
     inFile >> c; inFile >> cd; inFile >> cm;
     inFile >> jpos;
     ipos = (64+jpos)%97;
     pi97 = &u[ipos];
     pj97 = &u[jpos];
   }
}

void HepJamesRandom::showStatus() const
{
   std::cout << std::endl;
   std::cout << "----- HepJamesRandom engine status -----" << std::endl;
   std::cout << " Initial seed = " << theSeed << std::endl;
   std::cout << " u[] = ";
   for (HepInt i=0; i<97; ++i)
     std::cout << u[i] << " ";
   std::cout << std::endl;
   std::cout << " c = " << c << ", cd = " << cd << ", cm = " << cm
	     << std::endl;
   std::cout << " pi97 = " << pi97 << ", *pi97 = " << *pi97 << std::endl;
   std::cout << " pj97 = " << pj97 << ", *pj97 = " << *pj97 << std::endl;
   std::cout << "----------------------------------------" << std::endl;
}

void HepJamesRandom::setSeed(long seed, HepInt)
{
  // The input value for "seed" should be within the range [0,900000000]

  HepInt m, n;
  HepFloat s, t;
  long mm;

  long ij = seed/30082;
  long kl = seed - 30082*ij;
  long i = (ij/177) % 177 + 2;
  long j = ij % 177 + 2;
  long k = (kl/169) % 178 + 1;
  long l = kl % 169;

  theSeed = seed;

  for ( n = 1 ; n < 98 ; n++ ) {
    s = 0.0;
    t = 0.5;
    for ( m = 1 ; m < 25 ; m++) {
      mm = ( ( (i*j) % 179 ) * k ) % 179;
      i = j;
      j = k;
      k = mm;
      l = ( 53 * l + 1 ) % 169;
      if ( (l*mm % 64 ) >= 32 )
        s += t;
      t *= 0.5;
    }
    u[n-1] = s;
  }
  c = 362436.0 / 16777216.0;
  cd = 7654321.0 / 16777216.0;
  cm = 16777213.0 / 16777216.0;
  pi97 = &u[96];
  pj97 = &u[32];
}

void HepJamesRandom::setSeeds(const long* seeds, HepInt)
{
  setSeed(seeds ? *seeds : 19780503L, 0);
  theSeeds = seeds;
}

HepDouble HepJamesRandom::flat()
{
   HepDouble uni;

   do {
      uni = *pi97 - *pj97;
      if ( uni < 0.0 ) uni++;
      *pi97 = uni;
      
      if (pi97 == u) pi97 += 96;
      else pi97--;
      
      if (pj97 == u) pj97 += 96;
      else pj97--;
      
      c -= cd;
      if (c < 0.0) c += cm;
      
      uni -= c;
      if (uni < 0.0) uni += 1.0;
   } while ( uni <= 0.0 || uni >= 1.0 );
   
   return uni;
}

void HepJamesRandom::flatArray(const HepInt size, HepDouble* vect)
{
   HepDouble uni;
   HepInt i;

   for (i=0; i<size; ++i) {
      do {
         uni = *pi97 - *pj97;
         if ( uni < 0.0 ) uni++;
         *pi97 = uni;

         if (pi97 == u) pi97 += 96;
         else pi97--;

         if (pj97 == u) pj97 += 96;
         else pj97--;

         c -= cd;
         if (c < 0.0) c += cm;

         uni -= c;
         if (uni < 0.0) uni += 1.0;
      } while ( uni <= 0.0 || uni >= 1.0 );
      vect[i] = uni;
   }   
}

HepJamesRandom::operator unsigned int() {
   return (unsigned int)(flat() * exponent_bit_32) & 0xffffffff   |
         ((unsigned int)( *pi97 * exponent_bit_32)>>16)  & 0xff;
}

#if 0
std::ostream & operator<< ( std::ostream& os, const HepJamesRandom& e )
{
   HepString beginMarker = "JamesRandom-begin";
   HepString endMarker = "JamesRandom-end";

   HepInt pos = HepInt(e.pj97-e.u);
   os << " " << beginMarker << " ";
   os <<  e.theSeed << " ";
   for (HepInt i=0; i<97; ++i) {
     os << std::setprecision(20) << e.u[i] << " ";
   }
   os << std::setprecision(20) << e.c << " ";
   os << std::setprecision(20) << e.cd << " ";
   os << std::setprecision(20) << e.cm << " ";
   os << pos << " ";
   os << endMarker << " ";
   return os;
}
#endif

#if 0
std::istream & operator >> ( std::istream& is, HepJamesRandom& e )
{
   HepInt ipos, jpos;
   HepString beginMarker = " ";
   HepString endMarker = " ";

   is >> std::ws >> beginMarker;
   if( beginMarker != "JamesRandom-begin" ) {
     is.clear(std::ios::badbit | is.rdstate());
     std::cerr << "\nInput stream mispositioned or"
	       << "\nJamesRandom state description missing or"
	       << "\nwrong engine type found." << std::endl;
     return is;
   }
   is >> e.theSeed;
   for (HepInt i=0; i<97; ++i) {
     is >> e.u[i];
   }
   is >> e.c; is >> e.cd; is >> e.cm;
   is >> jpos;
   is >> std::ws >> endMarker;
   if( endMarker != "JamesRandom-end" ) {
     is.clear(std::ios::badbit | is.rdstate());
     std::cerr << "\nJamesRandom state description incomplete."
	       << "\nInput stream is probably mispositioned now." << std::endl;
     return is;
   }

   ipos = (64+jpos)%97;
   e.pi97 = &e.u[ipos];
   e.pj97 = &e.u[jpos];
   return is;
}

#endif
