// $Id: MTwistEngine.cpp,v 1.1.1.1 2000/04/03 21:03:45 wilko Exp $
// -*- C++ -*-
//
// -----------------------------------------------------------------------
//                             HEP Random
//                        --- MTwistEngine ---
//                      class implementation file
// -----------------------------------------------------------------------
// A "fast, compact, huge-period generator" based on M. Matsumoto and 
// T. Nishimura, "Mersenne Twister: A 623-dimensionally equidistributed 
// uniform pseudorandom number generator", to appear in ACM Trans. on
// Modeling and Computer Simulation.  It is a twisted GFSR generator
// with a Mersenne-prime period of 2^19937-1, uniform on open interval (0,1)
// =======================================================================
// Ken Smith      - Started initial draft:                    14th Jul 1998
//                - Optimized to get pow() out of flat() method
//                - Added conversion operators:                6th Aug 1998
// J. Marraffino  - Added some explicit casts to deal with
//                  machines where sizeof(int) != sizeof(long)  22 Aug 1998
// =======================================================================

#include "CLHEP/Random/Random.h"
#include "CLHEP/Random/MTwistEngine.h"
//#include "CLHEP/String/Strings.h"
#include <iomanip>

HepInt MTwistEngine::numEngines = 0;
HepInt MTwistEngine::maxIndex = 215;

MTwistEngine::MTwistEngine() 
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.))
{
  HepInt cycle = abs(HepInt(numEngines/maxIndex));
  HepInt curIndex = abs(HepInt(numEngines%maxIndex));
  long mask = ((cycle & 0x007fffff) << 8);
  long seedlist[2];
  HepRandom::getTheTableSeeds( seedlist, curIndex );
  seedlist[0] = (seedlist[0])^mask;
  seedlist[1] = 0;
  setSeeds( seedlist, 0 );
  count624=0;
  ++numEngines;
  for( HepInt i=0; i < 100; ++i ) flat();      // Warm up just a bit
}

MTwistEngine::MTwistEngine(long seed)  
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.))
{
  long seedlist[2]={seed,0};
  setSeeds( seedlist, 0 );
  count624=0;
  for( HepInt i=0; i < 100; ++i ) flat();      // Warm up just a bit
}

MTwistEngine::MTwistEngine(HepInt rowIndex, HepInt colIndex)  
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.))
{
  HepInt cycle = abs(HepInt(rowIndex/maxIndex));
  HepInt row = abs(HepInt(rowIndex%maxIndex));
  HepInt col = abs(HepInt(colIndex%2));
  long mask = (( cycle & 0x000007ff ) << 20 );
  long seedlist[2];
  HepRandom::getTheTableSeeds( seedlist, row );
  seedlist[0] = (seedlist[col])^mask;
  seedlist[1] = 0;
  setSeeds(seedlist, 0);
  count624=0;
  for( HepInt i=0; i < 100; ++i ) flat();      // Warm up just a bit
}

MTwistEngine::MTwistEngine( std::istream& is )  
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.))
{
  is >> *this;
}

MTwistEngine::~MTwistEngine() {}

MTwistEngine::MTwistEngine( const MTwistEngine & p )  
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.))
{
  *this = p;
}

MTwistEngine & MTwistEngine::operator=( const MTwistEngine & p ) {
  if( this != &p ) {
     for( HepInt i=0; i < 624; ++i )  mt[i] = p.mt[i];
     count624 = p.count624;
  }
  return *this;
}
HepDouble MTwistEngine::flat() {
  unsigned int y;

  if( count624 >= N ) {
    register HepInt i;

    for( i=0; i < NminusM; ++i ) {
      y = (mt[i] & 0x80000000) | (mt[i+1] & 0x7fffffff);
      mt[i] = mt[i+M]       ^ (y >> 1) ^ ((y & 0x1) ? 0x9908b0df : 0x0 );
    }

    for(    ; i < N-1    ; ++i ) {
      y = (mt[i] & 0x80000000) | (mt[i+1] & 0x7fffffff);
      mt[i] = mt[i-NminusM] ^ (y >> 1) ^ ((y & 0x1) ? 0x9908b0df : 0x0 );
    }

    y = (mt[i] & 0x80000000) | (mt[0] & 0x7fffffff);
    mt[i] = mt[M-1] ^ (y >> 1) ^ ((y & 0x1) ? 0x9908b0df : 0x0 );

    count624 = 0;
  }

  y = mt[count624];
  y ^= ( y >> 11);
  y ^= ((y << 7 ) & 0x9d2c5680);
  y ^= ((y << 15) & 0xefc60000);
  y ^= ( y >> 18);

  return                      y * mantissa_bit_32 +    // Scale to range 
         (mt[count624++] >> 11) * mantissa_bit_53 +    // fill remaining bits
                mantissa_bit_54 - mantissa_bit_106;    // make sure non-zero
}

void MTwistEngine::flatArray( const HepInt size, HepDouble *vect ) {
  for( HepInt i=0; i < size; ++i) vect[i] = flat();
}

void MTwistEngine::setSeed(long seed, HepInt) {
  theSeed = seed ? seed : 4357;
  mt[0] = (unsigned int)theSeed;
  for( HepInt i=1; i < 624; ++i ) {
    mt[i] = (69069 * mt[i-1]) & 0xffffffff;
  }
}

void MTwistEngine::setSeeds(const long *seeds, HepInt) {
  setSeed( (*seeds ? *seeds : 4357), 0 );
  theSeeds = seeds;
}

void MTwistEngine::saveStatus( const char filename[] ) const
{
   std::ofstream outFile( filename, std::ios::out ) ;
   if (!outFile.bad()) {
     outFile << theSeed << std::endl;
     for (HepInt i=0; i<624; ++i) outFile <<std::setprecision(20) << mt[i] << " ";
     outFile << std::endl;
     outFile << count624 << std::endl;
   }
}

void MTwistEngine::restoreStatus( const char filename[] )
{
   std::ifstream inFile( filename, std::ios::in);

   if (!inFile.bad() && !inFile.eof()) {
     inFile >> theSeed;
     for (HepInt i=0; i<624; ++i) inFile >> mt[i];
     inFile >> count624;
   }
}

void MTwistEngine::showStatus() const
{
   std::cout << std::endl;
   std::cout << "--------- MTwist engine status ---------" << std::endl;
   std::cout << std::setprecision(20);
   std::cout << " Initial seed      = " << theSeed << std::endl;
   std::cout << " Current index     = " << count624 << std::endl;
   std::cout << " Array status mt[] = " << std::endl;
   for (HepInt i=0; i<624; i+=5) {
     std::cout << mt[i]   << " " << mt[i+1] << " " << mt[i+2] << " " 
	       << mt[i+3] << " " << mt[i+4] << std::endl;
   }
   std::cout << "----------------------------------------" << std::endl;
}

MTwistEngine::operator HepFloat() {
  unsigned int y;

  if( count624 >= N ) {
    register HepInt i;

    for( i=0; i < NminusM; ++i ) {
      y = (mt[i] & 0x80000000) | (mt[i+1] & 0x7fffffff);
      mt[i] = mt[i+M]       ^ (y >> 1) ^ ((y & 0x1) ? 0x9908b0df : 0x0 );
    }

    for(    ; i < N-1    ; ++i ) {
      y = (mt[i] & 0x80000000) | (mt[i+1] & 0x7fffffff);
      mt[i] = mt[i-NminusM] ^ (y >> 1) ^ ((y & 0x1) ? 0x9908b0df : 0x0 );
    }

    y = (mt[i] & 0x80000000) | (mt[0] & 0x7fffffff);
    mt[i] = mt[M-1] ^ (y >> 1) ^ ((y & 0x1) ? 0x9908b0df : 0x0 );

    count624 = 0;
  }

  y = mt[count624++];
  y ^= ( y >> 11);
  y ^= ((y << 7 ) & 0x9d2c5680);
  y ^= ((y << 15) & 0xefc60000);
  y ^= ( y >> 18);

  return (HepFloat)(y * mantissa_bit_32);
}

MTwistEngine::operator unsigned int() {
  unsigned int y;

  if( count624 >= N ) {
    register HepInt i;

    for( i=0; i < NminusM; ++i ) {
      y = (mt[i] & 0x80000000) | (mt[i+1] & 0x7fffffff);
      mt[i] = mt[i+M]       ^ (y >> 1) ^ ((y & 0x1) ? 0x9908b0df : 0x0 );
    }

    for(    ; i < N-1    ; ++i ) {
      y = (mt[i] & 0x80000000) | (mt[i+1] & 0x7fffffff);
      mt[i] = mt[i-NminusM] ^ (y >> 1) ^ ((y & 0x1) ? 0x9908b0df : 0x0 );
    }

    y = (mt[i] & 0x80000000) | (mt[0] & 0x7fffffff);
    mt[i] = mt[M-1] ^ (y >> 1) ^ ((y & 0x1) ? 0x9908b0df : 0x0 );

    count624 = 0;
  }

  y = mt[count624++];
  y ^= ( y >> 11);
  y ^= ((y << 7 ) & 0x9d2c5680);
  y ^= ((y << 15) & 0xefc60000);
  y ^= ( y >> 18);

  return y;
}
/*
std::ostream & operator << ( std::ostream& os, const MTwistEngine& e )
{
   HepString beginMarker = "MTwistEngine-begin";
   HepString endMarker = "MTwistEngine-end";

   os << " " << beginMarker << " ";
   os << e.theSeed << " ";
   for (HepInt i=0; i<624; ++i) {
     os << std::setprecision(20) << e.mt[i] << " ";
   }
   os << e.count624 << " ";
   os << endMarker << " ";
   return os;
}
*/
/*
std::istream & operator >> ( std::istream& is, MTwistEngine& e )
{
   HepString beginMarker = " ";
   HepString endMarker = " ";

   is >> std::ws >> beginMarker;
   if( beginMarker != "MTwistEngine-begin" ) {
     is.clear(std::ios::badbit | is.rdstate());
     std::cerr << "\nInput stream mispositioned or"
	       << "\nMTwistEngine state description missing or"
	       << "\nwrong engine type found." << std::endl;
     return is;
   }
   is >> e.theSeed;
   for (HepInt i=0; i<624; ++i)  is >> e.mt[i];
   is >> e.count624;
   is >> std::ws >> endMarker;
   if( endMarker != "MTwistEngine-end" ) {
     is.clear(std::ios::badbit | is.rdstate());
     std::cerr << "\nMTwistEngine state description incomplete."
	       << "\nInput stream is probably mispositioned now." << std::endl;
     return is;
   }
   return is;
}

*/
