// $Id: Ranlux64Engine.cpp,v 1.1.1.1 2000/04/03 21:03:46 wilko Exp $
// -*- C++ -*-
//
// -----------------------------------------------------------------------
//                             HEP Random
//                       --- Ranlux64Engine ---
//                      class implementation file
// -----------------------------------------------------------------------
// A double-precision implementation of the RanluxEngine generator as 
// decsribed by the notes of the original ranlux author (Martin Luscher)
// =======================================================================
// Ken Smith      - Initial draft: 14th Jul 1998
//                - Removed pow() from flat method 14th Jul 1998
//                - Added conversion operators:  6th Aug 1998
// =======================================================================

//#include "CLHEP/String/Strings.h"
#include "CLHEP/Random/Random.h"
#include "CLHEP/Random/Ranlux64Engine.h"
#include <iomanip>

// Number of instances with automatic seed selection
HepInt Ranlux64Engine::numEngines = 0;
// Maximum index into the seed table
HepInt Ranlux64Engine::maxIndex = 215;

Ranlux64Engine::Ranlux64Engine()
: mantissa_bit_31 ( pow(0.5,31.)  ),
  mantissa_bit_53 ( pow(0.5,53.)  ),
  mantissa_bit_54 ( pow(0.5,54.)  ),
  mantissa_bit_106( pow(0.5,106.) )
{
   luxury = 1;
   HepInt cycle    = abs(HepInt(numEngines/maxIndex));
   HepInt curIndex = abs(HepInt(numEngines%maxIndex));
   numEngines +=1;
   long mask = ((cycle & 0x007fffff) << 8);
   long seedlist[2];
   HepRandom::getTheTableSeeds( seedlist, curIndex );
   seedlist[0] ^= mask;
   seedlist[1] = 0;

   setSeeds(seedlist, luxury);
}


Ranlux64Engine::Ranlux64Engine(long seed, HepInt lux)
: mantissa_bit_31 ( pow(0.5,31.)  ),
  mantissa_bit_53 ( pow(0.5,53.)  ),
  mantissa_bit_54 ( pow(0.5,54.)  ),
  mantissa_bit_106( pow(0.5,106.) )
{
   luxury = lux;
   long seedlist[2]={seed,0};
   setSeeds(seedlist, luxury);
}

Ranlux64Engine::Ranlux64Engine(HepInt rowIndex, HepInt colIndex, HepInt lux)
: mantissa_bit_31 ( pow(0.5,31.)  ),
  mantissa_bit_53 ( pow(0.5,53.)  ),
  mantissa_bit_54 ( pow(0.5,54.)  ),
  mantissa_bit_106( pow(0.5,106.) )
{
   luxury = lux;
   HepInt cycle = abs(HepInt(rowIndex/maxIndex));
   HepInt   row = abs(HepInt(rowIndex%maxIndex));
   HepInt   col = abs(HepInt(colIndex%2));
   long mask = (( cycle & 0x000007ff ) << 20 );
   long seedlist[2]; 
   HepRandom::getTheTableSeeds( seedlist, row );
   seedlist[0] ^= mask;

   seedlist[1]= 0;
   setSeeds(seedlist, luxury);
}

Ranlux64Engine::Ranlux64Engine( std::istream& is )
: mantissa_bit_31 ( pow(0.5,31.)  ),
  mantissa_bit_53 ( pow(0.5,53.)  ),
  mantissa_bit_54 ( pow(0.5,54.)  ),
  mantissa_bit_106( pow(0.5,106.) )
{
  is >> *this;
}

Ranlux64Engine::~Ranlux64Engine() {}

Ranlux64Engine::Ranlux64Engine(const Ranlux64Engine &p)
: mantissa_bit_31 ( pow(0.5,31.)  ),
  mantissa_bit_53 ( pow(0.5,53.)  ),
  mantissa_bit_54 ( pow(0.5,54.)  ),
  mantissa_bit_106( pow(0.5,106.) )
{
  *this = p;
}

Ranlux64Engine & Ranlux64Engine::operator=( const Ranlux64Engine &p ) {
  if (this != &p ) {
    theSeed = p.theSeed;
    theSeeds = p.theSeeds;
    for (HepInt i=0; i<12; ++i)
      seed_table[i] = p.seed_table[i];
    nskip = p.nskip;
    luxury = p.luxury;
    i_lag = p.i_lag;  j_lag = p.j_lag;
    carry = p.carry;
    count12 = p.count12;
  }
  return *this;
}

HepDouble Ranlux64Engine::flat() {

  HepDouble uni = seed_table[j_lag] - seed_table[i_lag] - carry;
  if(uni < 0. ){
     uni += 1.0;
     carry = mantissa_bit_54;
  }else{
     carry = 0.;
  }

  seed_table[i_lag] = uni;
  ++i_lag;
  ++j_lag;
  if(i_lag > 11) i_lag = 0;
  if(j_lag > 11) j_lag = 0;

  HepDouble next_random = uni;
  ++count12;

// every 12th number generation, several random numbers are generated
// and wasted depending upon the luxury level.

  if(count12 == 12 ){
     count12 = 0;
     for( HepInt i = 0; i < nskip ; ++i){
         uni = seed_table[j_lag] - seed_table[i_lag] - carry;
         if(uni < 0. ){
            uni += 1.0;
            carry = mantissa_bit_54;
         }else{
            carry = 0.;
         }
         seed_table[i_lag] = uni;
         ++i_lag;
         ++j_lag;
         if(i_lag > 11) i_lag = 0;
         if(j_lag > 11) j_lag = 0;
      }
  }
  return next_random;
}


void Ranlux64Engine::flatArray(const HepInt size, HepDouble* vect) {
  for( HepInt i=0; i < size; ++i ) vect[i] = flat(); 
}

void Ranlux64Engine::setSeed(long seed, HepInt lux) {

// The initialization is carried out using a Multiplicative
// Congruential generator using formula constants of L'Ecuyer
// as described in "A review of pseudorandom number generators"
// (Fred James) published in Computer Physics Communications 60 (1990)
// pages 329-344

  const HepInt ecuyer_a(53668);
  const HepInt ecuyer_b(40014);
  const HepInt ecuyer_c(12211);
  const HepInt ecuyer_d(2147483563);

  const HepInt lux_levels[3] = {109, 202, 397};
  theSeed = seed;

  if( (lux > 2)||(lux < 0) ){
     nskip = (lux >= 12) ? (lux-12) : lux_levels[1];
  }else{
     luxury = lux;
     nskip = lux_levels[luxury];
  }

  long init_table[24];
  long next_seed = seed;
  long k_multiple;
  HepInt i;
  
  for(i = 0;i != 24;i++){
     k_multiple = next_seed / ecuyer_a;
     next_seed = ecuyer_b * (next_seed - k_multiple * ecuyer_a)
                                       - k_multiple * ecuyer_c;
     if(next_seed < 0) next_seed += ecuyer_d;
     init_table[i] = next_seed & 0xffffffff;
  }    

  for(i = 0;i < 12; i++){
     seed_table[i] = (init_table[2*i  ]      ) * mantissa_bit_31 +
                     (init_table[2*i+1] >> 10) * mantissa_bit_53 +
                               mantissa_bit_54 - mantissa_bit_106;
  }

  if( seed_table[11] == 0. ) carry = mantissa_bit_54;
  
  i_lag = 11;
  j_lag = 7;
  carry = 0.0;
  count12 = 0;
}

void Ranlux64Engine::setSeeds(const long * seeds, HepInt lux) {
  setSeed( *seeds ? *seeds : 32767, lux );
  theSeeds = seeds;
}

void Ranlux64Engine::saveStatus( const char filename[] ) const
{
   std::ofstream outFile( filename, std::ios::out ) ;
   if (!outFile.bad()) {
     outFile << theSeed << std::endl;
     for (HepInt i=0; i<12; ++i)
       outFile <<std::setprecision(20) << seed_table[i] << " ";
     outFile << std::endl;
     outFile << i_lag << " " << j_lag << std::endl;
     outFile << std::setprecision(20) << carry << " " << count12 << std::endl;
     outFile << luxury << " " << nskip << std::endl;
   }
}

void Ranlux64Engine::restoreStatus( const char filename[] )
{
   std::ifstream inFile( filename, std::ios::in);

   if (!inFile.bad() && !inFile.eof()) {
     inFile >> theSeed;
     for (HepInt i=0; i<12; ++i)
       inFile >> seed_table[i];
     inFile >> i_lag; inFile >> j_lag;
     inFile >> carry; inFile >> count12;
     inFile >> luxury; inFile >> nskip;
   }
}

void Ranlux64Engine::showStatus() const
{
   std::cout << std::endl;
   std::cout << "--------- Ranlux engine status ---------" << std::endl;
   std::cout << " Initial seed = " << theSeed << std::endl;
   std::cout << " seed_table[] = ";
   for (HepInt i=0; i<12; ++i)
     std::cout << seed_table[i] << " ";
   std::cout << std::endl;
   std::cout << " i_lag = " << i_lag << ", j_lag = " << j_lag << std::endl;
   std::cout << " carry = " << carry << ", count12 = " << count12 << std::endl;
   std::cout << " luxury = " << luxury << " nskip = " << nskip << std::endl;
   std::cout << "----------------------------------------" << std::endl;
}
/*
std::ostream & operator << ( std::ostream& os, const Ranlux64Engine& e )
{
   HepString beginMarker = "Ranlux64Engine-begin";
   HepString endMarker = "Ranlux64Engine-end";

   os << " " << beginMarker << " ";
   os << e.theSeed << " ";
   for (HepInt i=0; i<12; ++i) {
     os << std::setprecision(20) << e.seed_table[i] << " ";
   }
   os << e.i_lag << " " << e.j_lag << " ";
   os << std::setprecision(20) << e.carry << " " << e.count12 << " ";
   os << e.luxury << " " << e.nskip << " ";
   os << endMarker << " ";
   return os;
}
*/
/*
std::istream & operator >> ( std::istream& is, Ranlux64Engine& e )
{
   HepString beginMarker = " ";
   HepString endMarker = " ";

   is >> std::ws >> beginMarker;
   if( beginMarker != "Ranlux64Engine-begin" ) {
     is.clear(std::ios::badbit | is.rdstate());
     std::cerr << "\nInput stream mispositioned or"
	       << "\nRanlux64Engine state description missing or"
	       << "\nwrong engine type found." << std::endl;
     return is;
   }
   is >> e.theSeed;
   for (HepInt i=0; i<12; ++i) {
     is >> e.seed_table[i];
   }
   is >> e.i_lag; is >> e.j_lag;
   is >> e.carry; is >> e.count12;
   is >> e.luxury; is >> e.nskip;
   is >> std::ws >> endMarker;
   if( endMarker != "Ranlux64Engine-end" ) {
     is.clear(std::ios::badbit | is.rdstate());
     std::cerr << "\nRanlux64Engine state description incomplete."
	       << "\nInput stream is probably mispositioned now." << std::endl;
     return is;
   }
   return is;
}
*/
