#ifndef DATAMANAGER_H
#define DATAMANAGER_H

//----------------------------------------------
//             J.A Hernando, Santa Cruz 02/29/00
//----------------------------------------------

#include <vector>
#include <string>
#include <iostream>

#include "Event/dataEventServer.h"
#include "Event/dataDetectorServer.h"
#include "Event/dataIOROOTServer.h"

typedef dataIOROOTServer dataIOServer;

//----------------------------------------
//! It stores, serves and updates the data
/*! 
<ul>
	<li> dataManager has/returns pointers to the following data servers:
	<ul>
		<li> dataDetectorServer, with geometrycal and calibration transient data.
		(\c det() method returns a pointer to dataDetectorServer).
		<li> dataEventServer, with event transient data. \c evt() method
		retruns a pointer to dataEventServer.
		<li> dataIOServer, with the persistent data. \c IO() method returns
		a pointer to the persistent data.
	</ul>
	<li> Any transient data can be recovered using the \c getData() method.
	<li> the \c update() method, update any transient data.
	<li> The persistent server could be replaced. The user should the typedef dataIOServer 
	to her/his defined <b>persistent data server </b> (i.e for ROOT, we use dataIOROOTServer).
</ul>
*/
//----------------------------------------
// JA Hernando, Santa Cruz, CA, 02/15/00
//----------------------------------------


//#############################
class dataManager : public serviceI
//#############################
{
public:

	//! constructor - construct the servers
	dataManager();
	//! destructor
	virtual ~dataManager();

	//! Singleton: pointer to itself 
	static dataManager* instance() {return m_instance;}

    //! returns the class derived from trsDataVI
	/*! The pointer to a dummy class of the derived type is needed in
	order to identify the template. JAH errors with dinamic_cast.*/
	template<class T> T* getData(std::string name, const T* dummy = 0) const { 
		T* t = static_cast<T*>(getTrsData(name));
		return t;
	}

	//! returns pointer to te dataEventServer
	inline dataEventServer*  evt()  const {return m_evt;}
	//! returns pointer to the dataDetectorServer
	inline dataDetectorServer*    det()  const {return m_det;}
	//! returns pointer to the persistency data
	inline dataIOServer*   IO()   const {return m_IO;}

	//! returns the pointer to the detector geometry in dataDetectorServer
	inline detectorGeo* geo()        const {return det()->geo();}

	//! write out information of the data servers
	virtual void writeOut() const;

protected:

	//! returns the trsData with that name in the servers
	virtual trsDataVI* getTrsData(std::string name) const;

private:

	//! Singleton: pointer to itself
	static dataManager* m_instance;

	//! pointer to event data
	dataEventServer* m_evt;
	//! pointer to detector data
	dataDetectorServer* m_det;
	//! pointer to persistent data
	dataIOServer* m_IO;
};

#endif
