#include "Event/processManager.h"
#include "Event/centellaAlgorithms.h"
#include "Event/userAlgorithms.h"
#include "Event/messageManager.h"

processManager* processManager::m_instance = 0;

//#####################################
processManager::processManager() 
//#####################################
{
	if (m_instance != 0) return;

	/*! create an algorithm server */
	m_algorithms     = new serverVI<algorithmVI>();
	/*! create an eventTask server */
	m_eventTasks = new serverVI<eventTask>();
	/*! create an algorithmTask server */
	m_algorithmTasks = new serverVI<algorithmTask>();
	/*! create an algConditionalTask server */
	m_algConditionalTasks = new serverVI<algConditionalTask>();

	/*! Always create a "iniOfRun" algorithmTask to execute the ini of run algorithms*/
	addAlgorithmTask("iniOfRun", new algorithmTask("iniOfRun"));
	/*! Always create a "Event" eventTask, which is the main
	algorithm to run all the process of on event*/
	addEventTask("runEvent", new eventTask("runEvent"));
	/*! Always create a "endOfRun" algorithmTask for execute the end of run algorithms*/
	addAlgorithmTask("endOfRun", new algorithmTask("endOfRun") );

	/*! default name of the application */
	m_applicationName = "centellaAlg";

	/*! dummy variables to set options */
	m_eventTaskName = "";
	m_algorithmTaskName = "";

	m_instance = this;

	/*! get the services */
	setName("processManager");
	/*! create and add into the server the system algorithms */
	centellaAlgorithms();
	/*! create and add into the server the user defined algorithms */
	userAlgorithms();
}
//#####################################
void processManager::setOption(std::string type, std::string name)
//#####################################
{
	/*! always call the primitive method to set the option into the optionVI servers */
	optionVI::setOption(type,name);
	if (type == "newEventTask") {
		/*! add a new eventTask into the server */
		addEventTask(name, new eventTask(name));
		message(" processManager: create eventTask "+name);
	} else if (type == "newAlgorithmTask") {
		/*! add a new algorithmTask into the server*/
		addAlgorithmTask(name, new algorithmTask(name));
		message(" created algorithmTask : "+name);
	} else if (type == "newAlgConditionalTask") {
		/*! add a new algorithmTask into the server*/
		addAlgConditionalTask(name, new algConditionalTask(name));
		message(" created algConditialTask : "+name);
	}
}
//#####################################
processManager::~processManager()
//#####################################
{
    // Free dynamically allocated algorithms
    // Note that m_algorithms contains each instance
    // of m_eventTasks, m_algorithmTasks, m_algConditionalTasks
    // So we don't free the elements of these vectors.
    unsigned int i;
    unsigned int size = m_algorithms->size();
    for (i = 0; i < size; i++) {
        if(m_algorithms->get(i)) delete (m_algorithms->get(i));
     }

	delete m_algorithms;
	delete m_eventTasks;
	delete m_algorithmTasks;
	delete m_algConditionalTasks;
}
//#####################################
void processManager::addAlgorithm(std::string name, algorithmVI* alg) 
//#####################################
{
	if (!m_algorithms->search(name)) m_algorithms->add(name,alg);
	else message(" algorithm already included in server "+name);
}
//#####################################
void processManager::addAlgorithmTask(std::string name, algorithmTask* alg)
//#####################################
{
	if (!m_algorithms->search(name)) {
		m_algorithmTasks->add(name,alg);
		addAlgorithm(name,alg);
	} else message(" algorithmTask already included in server "+name);
}
//#####################################
void processManager::addEventTask(std::string name, eventTask* evt)
//#####################################
{
	if (!m_algorithms->search(name)) {
		m_eventTasks->add(name,evt);
		addAlgorithm(name,evt);
	} else message(" not found evtProcess "+name,"GENERAL");
}
//#####################################
void processManager::addAlgConditionalTask(std::string name, algConditionalTask* alg)
//#####################################
{
	if (!m_algorithms->search(name)) {
		m_algConditionalTasks->add(name,alg);
		addAlgorithm(name,alg);
	} else message(" not found algConditionalTask "+name,"GENERAL");
}
//#####################################
algorithmVI* processManager::getAlgorithm(std::string name) const
//#####################################
{
	/*! search in algorithms. Note that algorithmTask, eventTask and
	algConditionalTask should be already included into the algorithm server */
	if (m_algorithms->search(name)) return m_algorithms->get(name);
	else message(" not found algorithm "+name,"GENERAL");
	return 0;
}
//#####################################
eventTask* processManager::getEventTask(std::string name) const
//#####################################
{
	if (m_eventTasks->search(name)) return m_eventTasks->get(name);
	message(" not found evtProcess "+name,"GENERAL");
	return 0;
}
//#####################################
algorithmTask* processManager::getAlgorithmTask(std::string name) const
//#####################################
{
	if (m_algorithmTasks->search(name)) return m_algorithmTasks->get(name);
	message(" not found algorithmTask "+name,"GENERAL");
	return 0;
}
//#####################################
algConditionalTask* processManager::getAlgConditionalTask(std::string name) const
//#####################################
{
	if (m_algConditionalTasks->search(name)) return m_algConditionalTasks->get(name);
	message(" not found algConditionalTask "+name,"GENERAL");
	return 0;
}
//########################################
void processManager::defineOption()
//########################################
{
	/*! define "eventTask" as an option to create a new eventTask*/
	optionVI::defineOption("newEventTask",&m_eventTaskName);
	/*! define "algorithmTask" as an option to create a new algorithmTask */
	optionVI::defineOption("newAlgorithmTask",&m_algorithmTaskName);
	/*! define "algorithmTask" as an option to create a new algorithmTask */
	optionVI::defineOption("newAlgConditionalTask",&m_algConditionalTaskName);
	/*! define "application" to set the name of the application algoritm */
	optionVI::defineOption("application",&m_applicationName);
}
//########################################
void processManager::writeOut() const
//########################################
{
	if (!acceptLevel()) return;	
	std::ostream& out = messageManager::instance()->out();

	messageVI::writeOut();

	/*! write out the application name */
	out << " application = " << m_applicationName <<"\n";

	/*! write out the list of algorithms (system + user) */
	if (m_algorithms->size() > 0) {
		out << " -- Algorithms declared in algorithms server --" << "\n";
		out << m_algorithms->nameList();
	}
	/*! write out the list of eventTask */
	if (m_eventTasks->size() > 0) {
		out << " -- eventTasks declared in the eventTasks server --" << "\n";
		out << m_eventTasks->nameList();
	}
	/*! write out the list of algorithmTaks s */
	if (m_algorithmTasks->size() > 0) {
		out << " -- algorithmTask declared in the algorithmTask server --" << "\n";
		out << m_algorithmTasks->nameList();
	}
		
	/*! write out the list of algConditionalTaks s */
	if (m_algConditionalTasks->size() > 0) {
		out << " -- algConditionalTask declared in the algConditionalTask server --" << "\n";
		out << m_algConditionalTasks->nameList(); 
	}

	/*! write out information about the eventTask */
	for (int ievt = 0; ievt < m_eventTasks->size(); ievt++ )
		m_eventTasks->get(ievt)->writeOut();

	/*! write out information about the algorithmTask */
	for (int itask = 0; itask < m_algorithmTasks->size(); itask++ )
		m_algorithmTasks->get(itask)->writeOut();
	
	/*! write out information about the algConditionalTaks s */
	for (int ic = 0; ic < m_algConditionalTasks->size(); ic++ )
		m_algConditionalTasks->get(ic)->writeOut();
}
