#ifndef processManager_h
#define processManager_h

#include "Event/serviceI.h"
#include "Event/serverVI.h"

#include "Event/eventTask.h"
#include "Event/algorithmTask.h"
#include "Event/algConditionalTask.h"

//-----------------------------------------------
//! Stores, servers and runs the algorithms and the multialgorithms
/*!
\e ProcessManager stores the system and user declared algorithms.<br>
\e ProcessManager returns the algorithms via the template method \c getAlgorithm(). <br> 
General features:	
<ul>
	<li> The <b> user algorithms </b> can declared in the class \ b userAlgorithms.
	<li> The user can add algorithms via the method \c addAlgorithm().
	<li> The method \c getAlgorithm() returns an algorithmVI called name from 
	the protected server of algorithms.
	<li> In addition, the  \c run(name) method will execute the run() method of the
	algorithm named. 
	<li> A particular algorithm can be called application. In priciple, to run
	a full program. \c runApplication() method will execute the application.
</ul>
Centella features:
<ul>
	<li> \b user can \e dinamically create multialgorithms and add them into
	the servers. The users can create algorithmTask, eventTask and algConditionalTask.
	All of them, are algorithms, and will be accesible via the \c getAlgorithm() method.
	<li> processManager has a protected server (serverVI) of eventTask. An eventTask
	is a multialgorithms with a collection of algorithms to be run for every event. The
	algorithms are clasified into generation, reconstruction and analysis parts.
	Proprely: an eventTask is an eventComposite algorithm with services provided by serviceI. These services
	An eventTask allows an external user (optionManager) to add algorithms to the generation,
	reconstruction and analysis part of the eventVI or add a new eventTask into another
	one.
	<li> processManager has a protected server (serverVI) of algorithmTask. An algorithmTask
	is a multialgorithm. Properly An algorithmTask
	is an algorithmComposite with system services provided by serviceI. 
	These services allow the possibility of an external user (most likely optionManager) 
	to add algorithms to the algorithmComposite (see algorithmTask for more information).
	<li> There is a fourth server with algConditionalTask. A conditionalTask is
	a angorithmConditional, an algorithm that runs if a condition of cutVI is fulfill.
	Properly: An algConditionalTask is an algorithmConditional with the services of serviceI. The user can create new algConditionalTask an add them into the server via
	the optionVI. Using the system services, an external user (again optionManager) 
	can set the algorithm to run and the cut that needs to be fulfill.
	<li> The \e Centella implementation of processManager has two already created algorithmTask
	named \e "iniOfRun" and \e "endOfRun", which should contain algorithms to be run at the
	start and end of the \b RUN.
	<li> This implementation also has a main eventTask called \e "runEvent" where 
	the eventTask defined by the user can be add to. 
	The "runEvent" is the process called by the runEventAlg algorithm that executes one event. 
	<li> Note that either eventTask, algorithmTask and algConditionalTask are algorithmVI themselves,
	and can be retrieved using the \c getAlgorithm() method.
</ul>
*/
//----------------------------------------------
//             J.A Hernando, Santa Cruz 02/29/00
//----------------------------------------------

//######################################
class processManager : public serviceI
//######################################
{
public:

	//! constructor - creates the algorithmUserServer
	processManager();
	//! destructor - delete the algorithmUserServer
	virtual ~processManager();
	
	//! Singleton, pointer to the class
	static processManager* instance() {return m_instance;}

	//! runs the application
	void runApplication() const {run(m_applicationName);}
	//! runs a given algorithm
	void run(std::string name) const {getAlgorithm(name)->run();}

	//! returns the required algorithm
	algorithmVI* getAlgorithm(std::string name) const;
	//! add algorithms into the server
	void addAlgorithm(std::string name, algorithmVI* alg);

	//! returns the required algorithmTask
	algorithmTask* getAlgorithmTask(std::string name) const;
	//! add an algorithmTask into the server
	void addAlgorithmTask(std::string name, algorithmTask* alg);

	//! returns the required eventTask
	eventTask* getEventTask(std::string name) const;
	//! add an eventTask into the server
	void addEventTask(std::string name, eventTask* evt);

	//! returns the required algConditionalTask
	algConditionalTask* getAlgConditionalTask(std::string name) const;
	//! add an algConditionalTask into the server
	void addAlgConditionalTask(std::string name, algConditionalTask* alg);

	//! write the information of this class
	virtual void writeOut() const;

protected:

	//! define the options 
	virtual void defineOption();
    //! create and include the evtProcess and endRunProcess that the User wants to execute
	virtual void setOption(std::string type, std::string name); //overwriten

private:
	
	//! Singleton: pointer to itself
	static processManager* m_instance;

	//! pointer to the server of algorithmVI s 
	serverVI<algorithmVI>*    m_algorithms;
	//! pointer to the eventTask Server
	serverVI<eventTask>*   m_eventTasks;
	//! pointer to the server of algorithmTask
	serverVI<algorithmTask>*  m_algorithmTasks;
	//! pointer to the server of algConditionalTask
	serverVI<algConditionalTask>* m_algConditionalTasks;

	//! name of the application
	std::string m_applicationName;

	//! dummy - name of the eventTask to create
	std::string m_eventTaskName;
	//! dummy - name of the algorithmTask to create
	std::string m_algorithmTaskName;
	//! dummy - name of the conditionalTask to create
	std::string m_algConditionalTaskName;
};

#endif
