///////////////////////////////////////////////////////////
//! This class processes Root Recon files 
/*!   This class is intended to provide useful manipulation
of the Root Event loop. Users put their analysis code
into the Go function (in MyRecon.C). They should not
need to look at MyRecon.h (except to see the interface)
It is targetted at the GLAST TB_Recon structure, which has
Recon as the main branch with sub-branches. This version 
handles the case of reading back from a file. 
allows for:
    init and re-init use of a Root file
    clear all histograms
    'go n events' allowing to continue on in the file
    or 'rewind'

    Example of use:

    gROOT->Macro("startmacro.c")  // load shared libs
    gROOT->LoadMacro("MyRecon.c");     // 'compile' class
    MyRecon* m = new MyRecon("MyRootFile.root"); // create MyRecon object
    m->Go(500);      // loop over 500 events. Go contains your analysis code
    ... look at histograms ...
    m->Go()          // look at remainder of file
    ... look at histograms ...
    m->HClr();      // clear histograms
    m->Init("AnotherRootFile.root");
    m->Go(50);
    ... and so on ...

After editing your Go function, you need to issue a gROOT->Reset() and
repeat the above sequence starting from the .L MyRecon.C.
*/
// Version 0.1 17-Mar-1999 Richard Creation
//////////////////////////////////////////////////////////


#ifndef MyRecon_h
#define MyRecon_h

class MyRecon {
   public :
       //! histogram file
       TFile*        histFile; 
       //! input file
       TFile*        f; 
       //! pointer to the analyzed TTree
       TTree          *fTree; 
       //! pointer to our current recon object
       Recon*        rec;

       //! default constructor
       MyRecon() {};
       //! constructor with root file name
       MyRecon(char* rootFileName); 
       //! default dtr
       ~MyRecon();  

       //! start next Go with this event
       void StartWithEvent(Int_t event);  
       //! re-init with this root file
       void Init(char* rootFileName); 
       //! Reset() all user histograms
       void HClr();
       //! delete all user histograms
       void AllHistDelete(); 
       //! define user histograms
       void HistDefine();   
       //! make list of user histograms
       void MakeHistList(); 
       //! reset for next Go to start at beginning of file
       void Rewind(); 
       //! loop over events
       void Go(Int_t numEvents=100000); 

 private:
     //! starting event
     Int_t m_StartEvent;
     //! list of user histograms
     TObjArray* HistList; 
};

#endif

// destructor: cleanup
MyRecon::~MyRecon() {

 histFile->Write();

 HClr();
 if (histFile) delete histFile;
 if (HistList) delete HistList;
 if (f) delete f;
 if (rec) delete rec;

}

MyRecon::MyRecon(char* rootFileName)
{
// constructor: set up root file and init things

   printf(" opening file %s \n",rootFileName);

   f = 0; histFile=0; fTree = 0; rec = 0;

   HistList = 0;
   HistDefine();
   MakeHistList();

   Init(rootFileName);


}

void MyRecon::StartWithEvent(Int_t event) {
// set a starting event for Go.
  m_StartEvent = event;
}

void MyRecon::Init(char* rootFileName)
{
// re-initialize file, tree, event. Histograms are *not* cleared.

//   Set branch addresses

       if (f) {
	 delete rec;
	 delete fTree;
	 delete f;
       }

      f = new TFile(rootFileName);
      fTree = (TTree*)gDirectory->Get("T");
      rec = new Recon();

      fTree->SetBranchAddress("Recon",&rec);

      m_StartEvent = 0;
}

void MyRecon::Rewind()
{
//   Start input file at beginning

       m_StartEvent = 0;

 return;
} 

void MyRecon::MakeHistList() {
// make a TObjArray of histograms

     if (HistList) delete HistList;

     TFile *histFile = (TFile*)gROOT->GetFile("Histograms.root");
     HistList = new TObjArray();

     TList* list = histFile->GetList();
     TIter*  iter = new TIter(list);

     TObject* obj = 0;

     while (obj=iter->Next()) {
       
       if (obj->InheritsFrom("TH1")) {
	 HistList->Add(obj);
       }
     }
     delete iter;
}

void MyRecon::HClr() {
// clear histograms

  if (!HistList) return;

  for (Int_t i=0; i < HistList->GetEntries(); i++) {
    TObject* obj = HistList->At(i);

    ((TH1*)HistList->At(i))->Reset();
  }

}

void MyRecon::AllHistDelete() {
// delete histograms

  if (HistList) {

    for (Int_t i=0; i < HistList->GetEntries(); i++) {
      TObject* obj = HistList->At(i);
      ((TH1*)obj)->SetDirectory(0);
      delete obj;
    }

    delete HistList;
  }
}

