// Project	: General Utilities Library
// Subproject	: Log
// File		: Messenger.cxx
// Author	: Patrick Hendriks
// Version	: 1.0
// Created	: 25-01-1997 (0.4)
// Last revision: 21-04-1998 (1.0)

#ifdef __GNUG__
#pragma implementation
#endif

#include "Messenger.h"



// *************************************************************************
// Global var.	: messenger
// Type		: Messenger
// Description	:
//	Common messenger object to be used by all clients.

Messenger
messenger = Messenger();


// *************************************************************************
// Class	: Messenger
// Static var.	: ms_enabled
// Access	: Protected, static
// Description	:
//	When it is 'true', messages are forwarded to the PrintControl.
//	Otherwise they are ignored.

bool
Messenger::ms_enabled = true;


// *************************************************************************
// Class	: Messenger
// Static var.	: ms_new_line
// Access	: Protected, static
// Description	:
//	True when the next output is printed on a new line.

bool
Messenger::ms_new_line = true;


// *************************************************************************
// Class	: Messenger
// Static var.	: ms_indentation
// Access	: Protected, static
// Description	:
//	The indentation level of the output.

unsigned char
Messenger::ms_indentation = 0;


// *************************************************************************
// Class	: Messenger
// Static var.	: ms_default_stream
// Access	: Private, static
// Description	:
//	The default stream for all messengers that do not have their own.

std::ostream*
Messenger::ms_default_stream = &std::cout;


// *************************************************************************
// Class	: Messenger
// Method	: Constructor
// Access	: Public
// Description	:
//	Create a new messenger that has no prefix

Messenger::Messenger() :
    m_prefix(), m_stream(&ms_default_stream)
{
}


// *************************************************************************
// Class	: Messenger
// Method	: Constructor
// Access	: Public
// Description	:
//	Create a new messenger object with the prefix of each line set
//	to 'prefix'.

Messenger::Messenger(const std::string& prefix) :
    m_prefix(prefix), m_stream(&ms_default_stream)
{
}


// *************************************************************************
// Class	: Messenger
// Method	: default_stream
// Access	: Public
// Description	:
//	Set the default stream for all messengers that do not have an
//	explicit output stream of themselves.

void
Messenger::default_stream(std::ostream& stream)
{
    ms_default_stream = &stream;
}


// *************************************************************************
// Class	: Messenger
// Method	: stream
// Access	: Public
// Description	:
//	Return the output stream of the messenger, which is Arve's
//	PrintControl when messages are enabled and a stream with a
//	null buffer when they are disabled. When m_new_line is true
//	the prefix is printed before the stream is returned.

std::ostream&
Messenger::stream()
{
    if (ms_enabled) {
	if (ms_new_line) {
	    for (int i = 0; i < ms_indentation; ++i) {
		**m_stream << "   ";
	    }
	    **m_stream << m_prefix;
	    ms_new_line = false;
	}
	return **m_stream;
    } else {
	static std::ostream* null_stream = new std::ostream(new NullBuf());
	return *null_stream;
    }
}

void
Messenger::stream(std::ostream& active_stream)
{
    static std::ostream* stream = 0;
    stream = &active_stream;
    m_stream = &stream;
}


// *************************************************************************
// Class	: Messenger
// Method	: operator<<
// Access	: Public
// Description	:
//	Executes the function 'p_func' with as an argument the stream of
//	'this'. This operator is e.g. used by 'endl', and therefore
//	'm_new_line' is set to true after the call to this function.

Messenger&
Messenger::operator<<(std::ostream& (*p_func)(std::ostream&))
{
    ((*p_func)(Messenger::stream()));
    ms_new_line = true;
    return *this;
}

