// Project	: General Utilities Library
// Subproject	: Log
// File		: Messenger.h
// Author	: Patrick Hendriks
// Version	: 1.0
// Created	: 25-01-1997 (0.4)
// Last revision: 21-04-1998 (1.0)
//
// Description	:
//	The Messenger class implements a simple logging scheme on an output
//	stream which is by default set to 'cout'.
//
//	The Messenger is by default enabled.
//
// Interface	:
//	Messenger()
//	    Create a new messenger that has no prefix.
//
//	explicit Messenger(const std::string& prefix)
//	    Create a new messenger object with the prefix of each line set
//	    to 'prefix'.
//
//	void prefix(const std::string& prefix)
//	    Set the prefix equal to 'prefix'.
//
//	void inc_indent(const unsigned char& step = 1)
//	    Increase the indentation by 'step'.
//
//	void dec_indent(const unsigned char& step = 1)
//	    Decrease the indentation by the maximum of step and the current
//	    indentation level.
//
//	void reset_indent()
//	    Rest the indentation level to 0.
//
//	static void default_stream(std::ostream& stream = std::cout)
//	    Set the default stream for all messengers that do not have an
//	    explicit output stream of themselves.
//
//	void stream(std::ostream& active_stream)
//	    Set the active stream to which the output is sent when messages
//	    are enabled.
//
//	virtual std::ostream& stream()
//	    Return the output stream of the messenger, which is the stream set
//	    by the user when messages are enabled and a stream with a
//	    null buffer when they are disabled. When m_new_line is true
//	    the prefix is printed before the stream is returned.
//
//	void enable()
//	    Enable the messengers.
//
//	void disable()
//	    Disable the messengers.
//
//	virtual Messenger& operator<<(std::ostream& (*p_func)(std::ostream&))
//	    Executes the function 'p_func' with as an argument the stream of
//	    'this'. This operator is e.g. used by 'endl', and therefore
//	    'm_new_line' is set to true after the call to this function.
//
// Global functions :
//	template <class T>
//	Messenger& operator<<(Messenger& m, const T& output)
//	    Print 'output' on the stream connected to 'm'.
//
// Global variables :
//	Messenger messenger
//	    Common messenger object to be used by all clients.

#ifndef UTIL_LOG_MESSENGER_H
#define UTIL_LOG_MESSENGER_H

#ifdef __GNUG__
#pragma interface
#endif

#include <iostream>
#include <string>



class Messenger
{
public:
    Messenger();
    explicit Messenger(const std::string& prefix);
    virtual ~Messenger() {} // because of virtual functions

    // Implicit copy constructor and assignment operator.

    // Maintenance
    void				prefix(const std::string& prefix);

    void				inc_indent(const unsigned char& step = 1);
    void				dec_indent(const unsigned char& step = 1);
    void				reset_indent();

    static void				default_stream(std::ostream& stream = std::cout);
    void				stream(std::ostream& active_stream);
    virtual std::ostream&		stream();

    // Execution
    void				enable();
    void				disable();

    virtual Messenger&			operator<<(std::ostream& (*p_func)(std::ostream&));
protected:
    class NullBuf: public std::streambuf					// Streambuf that eats all it is fed
    {
    };


    static bool				ms_enabled;
    static bool				ms_new_line;
    static unsigned char		ms_indentation;
    std::string				m_prefix;
private:
    static std::ostream*		ms_default_stream;
    std::ostream**			m_stream;
};


// *************************************************************************
// Inline functions :

inline void
Messenger::prefix(const std::string& prefix)
{
    m_prefix = prefix;
}


inline void
Messenger::inc_indent(const unsigned char& step)
{
    ms_indentation += step;
}


inline void
Messenger::dec_indent(const unsigned char& step)
{
    ms_indentation -= (step > ms_indentation ? ms_indentation : step);
}


inline void
Messenger::reset_indent()
{
    ms_indentation = 0;
}


inline void
Messenger::enable()
{
    ms_enabled = true;
}


inline void
Messenger::disable()
{
    ms_enabled = false;
}


// *************************************************************************
// Global variables:

extern Messenger messenger;


// *************************************************************************
// Global functions :

// TODO: Move back into util namespace when aCC support it!
template <class T>
inline Messenger&
operator<<(Messenger& m, const T& output)
{
    m.stream() << output;
    return m;
}


#endif	// UTIL_LOG_MESSENGER_H
