//     $Id: DisplayControl.cxx,v 1.3 2000/05/11 18:58:35 hernando Exp $
// Project: Arve graphics
//  Author: Toby Burnett
//
// implementation of  Display control

#ifdef __GNUG__
#pragma implementation
#endif

#include "DisplayControl.h"
#include "graphics/GUI.h"   // for the graphics window

#include "graphics/ArvePlottableRep.h"
#include "graphics/SceneControl.h"
#include "graphics/ArveScene.h"
#include "graphics/Vrml.h"
#include "graphics/Draw2D.h"
#include "graphics/PSdraw.h"


#include "control/PrintControl.h"
#include "control/SimpleCommand.h"

#include <string>  //for stl string class
#include <fstream>
#include <strstream>


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
DisplayControl::DisplayControl(Menu& menu, SceneControl* control)
    : m_control(control),
     m_axis_size(100),
     m_running(false)
{
    m_user_menu =m_sub_menu = &menu.sub_menu("Display");

    // schedule to have our finishSetup called
    menu.add(new MenuClient<DisplayControl>(this));


    // get a pointer to the scene
    m_scene = &m_control->scene();
    m_2d_scene = &m_control->scene_2d();

    m_canvas = m_control->view().canvas();

    // give the scene control a pointer to the menu, so it can pop it up
    m_control->set_menu(m_sub_menu);

}
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
DisplayControl::~DisplayControl()
{
    delete m_control;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//     VRML_command class
class DisplayControl::VRML_command : public Command {
public:
    VRML_command(DisplayControl* dcon):m_dcon(dcon){}
    void execute(); 
private:
    DisplayControl* m_dcon;
};

void DisplayControl::VRML_command::execute() {
    static std::string filename("glast.wrl");
    char* answer = GUI::instance()->askForFileName(filename.c_str(),"save"
				,"vrml files (*.wrl)|*.wrl|all files|*.*|") ;
    if( !answer) return;	// check for cancel
    filename = answer;
    std::ofstream vrml_file(filename.c_str());
    if( vrml_file.rdbuf()->is_open() ) m_dcon->vrml(vrml_file);
    else GUI::instance()->inform("Could not open the file");
}

class DisplayControl::PrintInstructions : public Command {
    friend class DisplayControl;
    PrintInstructions(SceneControl* disp, std::ostream* out):m_disp(disp), m_out(out){}
    void execute(){m_disp->print_commands(*m_out);}
    SceneControl* m_disp;
    std::ostream* m_out;

};
//=====================================================================
//   Private class used by DisplayControl to define axes
class DisplayControl::Axes: public ArvePlottableRep
{
public:
    Axes(double scale=1.);
    // the constructor creates the representation

    // dummy  clear
    void clear(){};
    void update(){};
    
private:
};


DisplayControl::Axes::Axes(double scale)
{
    static Hep3Vector unitvector[3] =
    {Hep3Vector(1.,0.,0.),
    Hep3Vector(0.,1.,0.),
    Hep3Vector(0.,0.,1.)
    };
    
    static const char* axislabel[3]={"x","y","z"};
    for( int i=0; i<3; i++){
        move_to(Hep3Vector(0,0,0));
        line_to((scale)*unitvector[i]);
        drawText(axislabel[i]);
    }
    flush();
}
//=====================================================================
//   Private class used by DisplayControl to display reference point
class DisplayControl::ReferencePoint : public Arve2D
{
public:
    ReferencePoint(const DisplayControl& dc):m_dc(dc){};
    void draw(Draw2D* canvas) {
        std::strstream text;
        Hep3Vector p(m_dc.reference_point());
        text << "ref pt: " << p << '\0';
        canvas->draw_string(0, -0.9f, text.str());
        text.freeze(false);
    }
private:
    const DisplayControl& m_dc;
};

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void
DisplayControl::finishSetup()
//---------------------------
{

    m_sub_menu->add_separator();
    if( m_axis_size != 0)  add( new Axes(m_axis_size),"axes" );
    add( new ReferencePoint(*this),"reference point", false);

    m_sub_menu->add_toggle("3D Enabled", m_scene->enabled(),
	new SimpleCommand<ArveScene>(m_scene,&ArveScene::enable),
	new SimpleCommand<ArveScene>(m_scene,&ArveScene::disable)
		);
    m_sub_menu->add_toggle("2D Enabled", m_2d_scene->enabled(),
	new SimpleCommand<ArveScene>(m_2d_scene,&ArveScene::enable),
	new SimpleCommand<ArveScene>(m_2d_scene,&ArveScene::disable)
		);
    m_sub_menu->add_button("clear", 
	new SimpleCommand<DisplayControl>(this,&DisplayControl::clear));

    m_sub_menu->add_button("update", 
	new SimpleCommand<DisplayControl>(this,&DisplayControl::update));

    m_sub_menu->add_button("Dump to postscript",
	new SimpleCommand<DisplayControl>(this,&DisplayControl::postScript));
    m_sub_menu->add_button("Dump to vrml",new VRML_command(this));
    m_sub_menu->add_button("Print instructions",
			   new PrintInstructions(m_control,PrintControl::instance()->out()));

    m_running = true;
    m_scene->update();
    m_2d_scene->update();
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void DisplayControl::setAxisSize(float size)
{
    m_axis_size = size;
}
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void DisplayControl::redisplay()
//---------------------------
// This is necesary to prevent the setup calls for ShowIt and HideIt calling
// Motif before the display has been realized
{
    if( m_running )
	m_control->redisplay();
}
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void DisplayControl::clear()
//-------------------------
{
    m_scene->clear();
    m_2d_scene->clear();
    redisplay();
}
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void DisplayControl::update()
//-------------------------
{
    m_scene->update();
    m_2d_scene->update();
    redisplay();
}
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


///////////////////////////////////////////////////////////////////////////////
//   add a view to display, with a buttons or a toggle in the Display menu
/////////////////////////////////////////////////////////////////////////////////
// File-scope stuff to define hide and show commands
    class DisplayControl::ShowIt : public Command    {
    friend class DisplayControl;
	ShowIt( DisplayControl* s,ArvePlottableRep * rep, bool f)
	    :m_scene(s),m_rep(rep),m_flag(f){}
	void execute()
	{
	    m_rep->show(m_flag);
            if( m_rep->empty() ) m_rep->update();
	    m_scene->redisplay();
	}
	DisplayControl* m_scene;
	ArvePlottableRep* m_rep;
	bool m_flag;
    };
    class DisplayControl::HideIt : public Command    {
    friend class DisplayControl;
	HideIt( DisplayControl* s,ArvePlottableRep * rep, bool f)
	    :m_scene(s),m_rep(rep),m_flag(f){}
	void execute()
	{
	    m_rep->hide(m_flag);
	    m_scene->redisplay();
	}
	DisplayControl* m_scene;
	ArvePlottableRep* m_rep;
	bool m_flag;
    };

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void
DisplayControl::add(ArvePlottableRep* pRep, const std::string& name, int state)
{
    m_scene->add(pRep);
	
    // now control it, if requested
    if( !name.length() ) return;

    add_button(pRep,name,state);
}
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void
DisplayControl::add(Arve2D* pRep, const std::string& name, int state)
{
    m_control->scene_2d().add(pRep);

    if( name.empty() ) return;
    add_button(pRep, name, state);

}
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void DisplayControl::add_button(ArvePlottableRep* pRep, 
                                const std::string& name, int state)
{

    Command* showit = new ShowIt(this, pRep, state>=0);
    Command* hideit = new HideIt(this, pRep, state>=0);
    if( state <0 ) {
	// special flag to make separate show and hide buttons
	m_user_menu->add_button(std::string("show ")+name, showit);
	m_user_menu->add_button(std::string("hide ")+name, hideit);
    } else {
	if( state>0 ) showit->execute(); else hideit->execute();
	m_user_menu->add_toggle(name, state>0, showit, hideit );
    }

}
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void DisplayControl::useMenu(SubMenu* m)
{
    m_user_menu = m? m : m_sub_menu;
}
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void
DisplayControl::setTitle(const std::string& t){GUI::instance()->setTitle(t.c_str());}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
const Hep3Vector&  DisplayControl::reference_point()const
//--------------------------------------------------
{
  return m_control->reference_point();
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void DisplayControl::vrml(std::ostream& pout)
//--------------------------------------
// ask the scene to dump itself to a VRML file
{
  Vrml output(pout);
    ((ArveScene*)m_scene)->draw(&output);
}
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void DisplayControl::postScript()
//--------------------------------
// ask the scene to draw itself to a file in postscript
{
    static const char* filename = "temp.ps";
    PSdraw ps(filename);
    if( !ps() ) {
	GUI::instance()->inform("Cannot open file for postscript output");
	return;
    }

    m_control->redisplay(&ps);
    std::ostrstream msg; msg << "Postscript version written to file " << filename << '\0';
    GUI::instance()->inform(msg.str());
}









