//     $Id: GUI.h,v 1.3 2000/05/11 18:58:35 hernando Exp $
// Project: Atlas top-level prototype
//  Author: Toby Burnett

#ifndef __GUI_H
#define __GUI_H

#include "control/Command.h"

class SceneControl;
#include <iostream>

using namespace std;

class GUIostream : public std::ostream
// define a special abstract stream for writing to a text window, 
// presumably a scrollable editor
// sub class can implement the flush and clear functions
{
  public:
    virtual ~GUIostream(){};
    virtual void clear(){};
    virtual void flush(){};
//THB??    GUIostream(std::streambuf* s):std::ostream(s){};
        GUIostream(streambuf* s):ostream(s){};
};


class GUI {
// Abstract base class for  GUI, which defines a factory method, createGUI for
// the "real" GUI.
// Defines methods to set the window title, to define a menu bar with pull-down command menus,
// (call-backs to Command objects), to use a graphical display (manipulated by an appropriate
// SceneControl object), and to direct printout to a separate window.
public:
    virtual ~GUI();

    virtual void addToMenu(const char* title, Command* command);
    // Add a button to the current menu, either the menu bar, or a pull-down menu

    virtual void addToggleToMenu(const char* title,int state, Command* set, Command* unset);
    // Add a toggle button to the current menu. state is the initial state (on or off)

    class Menu; // empty class to tag menu bar or sub menu

    virtual Menu* beginPullDownMenu(const char* title, Menu* m=0);
    // start adding commands to a pull-down menu, which may be a submenu of a previous one.
    // If zero, add to menubar.
    // Return tag

    virtual void restorePullDownMenu(Menu* m);
    // Restore a previously-defined pull-down menu

    virtual void endPullDownMenu();
    // Stop adding commands to a pull-down menu and revert to the menu bar

    virtual void menuSeparator();
    // Insert a separator bar in the current pull-down menu

    virtual SceneControl* graphicsWindow(float size=100)=0;
    // Create and return a SceneControl object that can create Arve graphics in the graphical
    // portion of the window.

    virtual GUIostream* textWindow(const char*)=0;
    // Create and return an ostream object. (Which may just be cout for unix.)

    virtual void run();
    // Start the window system. For a real GUI, this is a message loop.

    virtual void quit();
    // End the window system. Typically bound to a "Exit" button.

    virtual void setTitle(const char* newtitle);
    // Set the window title

    virtual void processMessages();
    // Process waiting messages (needed if single thread)

    virtual char * askForFileName(const char* a, const char* b, const char* c);
    // File name dialog box.

    virtual char * askUser(const char* promptString,const char* defaultString);
    // Simple dialog box that sets a string

    virtual void inform(const char* msg);
    // Dialog box with a message.

    static GUI* createGUI(const char* appname=0, const char* title=0);
    // Return a GUI. Expect subclass to implement. Prevents more that one GUI implementation
    // in an executable

    static GUI* instance();
    static bool running;  // use to check that GUI has been started

protected:
    GUI(const char* title);  // abstract class
    const char* _winTitle;  // window title

};

inline GUI::GUI(const char* title):_winTitle(title){}  // abstract class
inline GUI::~GUI(){}
inline void    GUI::addToMenu(const char* , Command* ){}
inline void    GUI::addToggleToMenu(const char*, int, Command*, Command* ){}
inline GUI::Menu*   GUI::beginPullDownMenu(const char*, GUI::Menu* m){return 0;}
inline void    GUI::restorePullDownMenu(GUI::Menu* ){}
inline void    GUI::endPullDownMenu(){}
inline void    GUI::menuSeparator(){}
inline void    GUI::run(){}
inline void    GUI::quit(){}
inline void    GUI::setTitle(const char*){}
inline void    GUI::processMessages(){}
inline char *  GUI::askForFileName(const char*, const char*, const char* ){return 0;}
inline char *  GUI::askUser(const char* ,const char* ){return 0;}
inline void    GUI::inform(const char* ){}


class GUICommand : public Command {
// Special form of SimpleCommand to create Command objects bound to GUI methods
public:
   typedef void (GUI::* Action)();

   GUICommand(GUI* r, Action a):
    _receiver(r), _action(a) {}

   virtual void execute(){ (_receiver->*_action)();}
private:
   GUI* _receiver;
   Action _action;
};

class GUIostreamCommand : public Command {
// Special form of SimpleCommand to create Command objects bound to GUIostream methods
public:
   typedef void (GUIostream::* Action)();

   GUIostreamCommand(GUIostream* r, Action a):
    _receiver(r), _action(a) {}

   virtual void execute(){ (_receiver->*_action)();}
private:
   GUIostream* _receiver;
   Action _action;
};

#endif

