
#include "instrument/calorimeterDetGeo.h"
#include "Event/messageManager.h"

//######################################################
calorimeterDetGeo::calorimeterDetGeo()
//######################################################
{
	createNav();
	createCal();

	orderLayers();

	m_nameOut = "LOG";
}
//################################################
calorimeterDetGeo::~calorimeterDetGeo()
//################################################
{
	int nEl = m_calorimeterNav->size();
	for (unsigned int i = 0; i < nEl; i++) delete m_calorimeterNav->el(i);
}
//################################################
void  calorimeterDetGeo::draw(ArvePlottableRep& v) 
//################################################
{
	int nlayers = m_calorimeterNav->size(m_nameOut);
	for (int ilayer= 0; ilayer < nlayers; ilayer ++) {
		m_calorimeterNav->el(m_nameOut,ilayer)->draw(v);
	}
}
//################################################
void  calorimeterDetGeo::writeOut() const
//################################################
{
	if (!messageManager::instance()->acceptLevel("DEBUG_MEDIUM")) return;

	std::ostream& out = messageManager::instance()->out();
	int nlayers = m_calorimeterNav->size(m_nameOut);
	for (int ilayer= 0; ilayer < nlayers; ilayer ++) {
		m_calorimeterNav->el(m_nameOut,ilayer)->writeOut();
	}
}
//--------------- private ----------------------
//##############################################
void calorimeterDetGeo::createNav()
//##############################################
{
	m_calorimeterNav = new navigator<detGeo*>(new detGeo(0,detGeo::PASSIVE,0),
		"CALORIMETER");
	m_calorimeterNav->add(new detGeo(0,detGeo::X,0),"XCALORIMETER");	
	m_calorimeterNav->add(new detGeo(0,detGeo::Y,0),"YCALORIMETER");
}
//##############################################
void calorimeterDetGeo::createCal()
//##############################################
{
	for (int iview = 0; iview < calorimeterGeo::numViews(); iview++) {
		detGeo::axis a = (iview == 0? detGeo::X : detGeo::Y);
		for (int ilayer = 0; ilayer < calorimeterGeo::numLayers(); ilayer ++) {
			createLayer(ilayer,a);
		}
	}
}
//##############################################
void calorimeterDetGeo::createLayer(int ilayer, detGeo::axis a)
//##############################################
{
	double xpos = 0.;
	double ypos = 0.;
	double zfar = 0.;
	if (a == detGeo::Y) zfar = layerHeight();
	double zpos = Z0()+2.*ilayer*layerHeight()+zfar;

	double xsize = calorimeterGeo::layerWidth();
	double ysize = calorimeterGeo::layerWidth();
	double zsize = calorimeterGeo::layerHeight();

	Point P(xpos,ypos,zpos);
	Point S(0.5*xsize,0.5*ysize,0.5*zsize);

	detGeo* layer = new detGeo(ilayer,a,ilayer,P,S);
	layer->setMaterial("CSI",0.);
	layer->setName("LAYER");
	getNavigator(a)->add(layer,"LAYER");
	for (int ilog = 0; ilog < calorimeterGeo::numLogs();  ilog ++) 
		createLog(ilayer,a,ilog);
}

//##############################################
void calorimeterDetGeo::createLog(int ilayer, detGeo::axis a, int ilog)
//##############################################
{
	detGeo* layer = getLayer(ilayer,a);
	double xpos = layer->position().x();
	double ypos = layer->position().y();
	double zpos = layer->position().z();
	
	int nlogs = calorimeterGeo::numLogs();
	double posRef = -0.5*(nlogs-1)*(calorimeterGeo::logWidth()+
		calorimeterGeo::logGap());
	double pos = posRef+ilog*(calorimeterGeo::logWidth()+
		calorimeterGeo::logGap());
	if (fabs(pos) < 1e-5) pos =0.; 
	if (a == detGeo::X) xpos = pos;
	else ypos = pos;

	double xsize = calorimeterGeo::logLength();
	double ysize = calorimeterGeo::logLength();
	if (a == detGeo::X) xsize = calorimeterGeo::logWidth();
	else ysize = calorimeterGeo::logWidth();
	double zsize = calorimeterGeo::logHeight();
	
	Point P(xpos,ypos,zpos);
	Point S(0.5*xsize,0.5*ysize,0.5*zsize);

	detGeo* log = new detGeo(ilayer,a,ilog,P,S);
	log->setMaterial("CSI",0.);
	log->setName("LOG");

	getNavigator(ilayer,a)->add(log,"LOG");
}
//######################################################
navigator<detGeo*>* calorimeterDetGeo::getNavigator(detGeo::axis a) const 
//######################################################
{ 
	if (a == detGeo::X) return m_calorimeterNav->nav("XCALORIMETER",0);
	else return m_calorimeterNav->nav("YCALORIMETER",0);
}
//######################################################
navigator<detGeo*>* calorimeterDetGeo::getNavigator(int ilayer, detGeo::axis a) const
//######################################################
{
	return getNavigator(a)->nav("LAYER",ilayer);
}
//######################################################
navigator<detGeo*>* calorimeterDetGeo::getNavigator(int ilayer, detGeo::axis a, int ilog) const
//######################################################;
{
	return getNavigator(ilayer,a)->nav("LOG",ilog);
}
//######################################################
void calorimeterDetGeo::orderLayers()
//######################################################
{
	int nLayers = m_calorimeterNav->size("LAYER");
	for (int ilayer =0; ilayer < nLayers; ilayer++) {
		double z0 = m_calorimeterNav->el("LAYER",ilayer)->position().z();
		int inext = -1;
		int iprevious = -1;
		double dmax = 1e6;
		double dmin = 1e6;
		for (int jlayer = 0; jlayer < nLayers; jlayer++) {
			if (jlayer != ilayer) { 
				double z = m_calorimeterNav->el("LAYER",jlayer)->position().z();
				if (z<z0 && fabs(z-z0)< dmin){
					iprevious = jlayer;
					dmin = fabs(z-z0);
				}
				if (z>z0 && fabs(z-z0)<dmax){
					inext = jlayer;
					dmax = fabs(z-z0);
				}
			}
		}
		if (inext>-1) m_calorimeterNav->nav("LAYER",ilayer)->
			setNext(m_calorimeterNav->nav("LAYER",inext));
		if (iprevious>-1) m_calorimeterNav->nav("LAYER",ilayer)->
			setPrev(m_calorimeterNav->nav("LAYER",iprevious));
	}

}
