
#include "reconstruction/CalClustersAlg.h"
#include "Event/dataManager.h"
#include "Event/messageManager.h"
#include "reconstruction/CsIClusters.h"
#include "reconstruction/CalRecLogs.h"
#include "reconstruction/gamma.h"
#include "reconstruction/Midnight.h"

int nbins;
std::vector<double> g_elayer;
double slope,size,radlen;

double fit_energy;
double ki2;
double fit_alpha;
double fit_lambda;
double fit_start;
double energy_err;
double alpha_err;
double lambda_err;
double start_err;


static double gam_prof(double *par, int i)
{
	double result =0; 

	double length = ((2.3*i+par[2])/1.85)/slope;

	// Evaluation of the parameters of CsI at this energy	

	double alpha = par[1];

	double lambda = par[3];

	double x=length/lambda;
	double dx = 2.3 / (1.85 *lambda)/slope;
	
	double gamma1 =0;
	double gamma2 = 0;

	// Now we will calculate the gamma incomplete function

	// gamma1 = integration from 0 to x	
	gamma1 = Gamma(alpha,x);
	x += dx;

	// gamma2 = integration from 0 to x+dx	
	gamma2 = Gamma(alpha,x);
	
	// the result of integration over Xtal pathlength is E*(gamma2 - gamma1)
	result = par[0]*(gamma2 - gamma1);
//	std::cout<< result << "\n";
	return result;
}



static void fcn(int &npar, double *gin, double &f, double *par, int iflag)
{
     int i;
	//calculate 'chisquare'
   double chisq = 0;
   double dlt;
   
   for (i=0;i<nbins; i++) {
     dlt  = (g_elayer[i]-gam_prof(par,i));
     if(g_elayer[i]>0.001) chisq += dlt*dlt/g_elayer[i];
   
   }
   f = chisq;
}


//################################################
void CalClustersAlg::Profile(double eTotal, CsICluster* cl)
//################################################
{
//	std::cout << "in profile " << eTotal << "\n";
	if( eTotal<500.) return;  // algorithm is useless under several GeV
	if( slope == 0) return; // protect against numerical errors.

	// Vector of step, initial min and max value
	float vstrt[5];
	float stp[5];
	float bmin[5];
	float bmax[5];

	// We are working in GeV
    double eTotal_GeV = eTotal / 1000;

	// par[0] is energy
	// par[1] is alpha
	// par[2] is the starting point
	// par[3] is lambda

	// Init start values and bounds

	// starting value of each parameter
	vstrt[0] = eTotal_GeV;
	vstrt[1] = 2.65 * exp(0.15*log(eTotal_GeV));  // parametrisation of alpha
	vstrt[2] = 1.8f;			// eq to 1X0 in CsI
	vstrt[3] = 2.29 * exp(-0.031*log(eTotal_GeV));
	// step value of each parameter
	stp[0] = 0.1f;
	stp[1] = 0.1f;
	stp[2] = 0.1f;
	stp[3] = 0.2f;

	// minimum value of each parameter
	bmin[0] = 0.5*eTotal_GeV;
	bmin[1] = 0.5;
	bmin[2] = -5;
	bmin[3] = 0.;

	// maximum value of each parameter
	bmax[0] = 5 * eTotal_GeV;
	bmax[1] = 15;
	bmax[2] = 10;
	bmax[3] = 10;

	// Those are the arguments for Minuit
    double arglist[10];
    int ierflg = 0;

	// Set no output because of tuple output
    arglist[0] = -1;
    minuit->mnexcm("SET PRI", arglist ,1,ierflg);
    
	// idem with warnings ( minuit prints warnings when the Hessian matrix is not positive )
    minuit->mnexcm("SET NOW", arglist ,1,ierflg);

	arglist[0] = 1;
    minuit->mnexcm("SET ERR", arglist ,1,ierflg);

	// defines the strategy used by minuit
	// 1 is standard
	arglist[0] = 2;
    minuit->mnexcm("SET STR", arglist ,1,ierflg);


	// Defines parameters
	
    minuit->mnparm(0, "Energy",    vstrt[0], stp[0], bmin[0],bmax[0],ierflg);
    minuit->mnparm(1, "Alpha",    vstrt[1], stp[1], bmin[1],bmax[1],ierflg);
    minuit->mnparm(2, "Starting point",  vstrt[2], stp[2], bmin[2],bmax[2],ierflg);
    minuit->mnparm(3, "Lambda",  vstrt[3], stp[3], bmin[3],bmax[3],ierflg);

    // Fix some parameters
	// alpha
    arglist[0] = 2;
    minuit->mnexcm("FIX ", arglist ,1,ierflg);

	// lambda
    arglist[0] = 4;
    minuit->mnexcm("FIX ", arglist ,1,ierflg);


   // Calls Migrad with 500 iterations maximum
   arglist[0]=500;
   arglist[1]=1;
   minuit->mnexcm("MIGRAD",arglist,2,ierflg);

   // Retrieve parameter information 

   minuit->GetParameter( 0, fit_energy,energy_err ); 
   minuit->GetParameter( 1, fit_alpha,alpha_err ); 
   minuit->GetParameter( 2, fit_start,start_err ); 
   minuit->GetParameter( 3, fit_lambda,lambda_err ); 

   arglist[0] = 3;
   minuit->mnexcm("CALL FCN", arglist ,1,ierflg);

   // Get chi-square
   double edm,errdef;
   int nvpar,nparx,icstat;
   minuit->mnstat(ki2,edm,errdef,nvpar,nparx,icstat);

   // Fills data
   cl->setFitEnergy(1000*fit_energy);    // Conversion to MeV
   cl->setProfChisq(ki2);
   cl->setCsiStart(fit_start);
   cl->setCsiAlpha(fit_alpha);
   cl->setCsiLambda(fit_lambda);
//   std::cout << "fit_ener " << fit_start << "\n";

	// Clear minuit
   arglist[0] = 1;
   minuit->mnexcm("CLEAR", arglist ,1,ierflg);

  

}


//################################################
CalClustersAlg::CalClustersAlg()
//################################################
{
	// Minuit object
	minuit = new Midnight(5);
	
	//Sets the function to be minimized
	minuit->SetFCN(fcn); 

	// initialisation of minuit
	minuit->mninit(16,17,18);

}

//################################################
CalClustersAlg::~CalClustersAlg()
//################################################
{
	// delete Minuit object
	delete minuit;

}

//################################################
void CalClustersAlg::initialize()
//################################################
{
	m_CsIClusterList = dataManager::instance()->getData("CsIClusterList",m_CsIClusterList);
	m_CalRecLogs  = dataManager::instance()->getData("CalRecLogs",m_CalRecLogs);
	m_CsIClusterList->clear();
	
	g_elayer.clear();
    
}
//################################################
void CalClustersAlg::execute()
//################################################
{
	int nLogs = m_CalRecLogs->num();
	double ene = 0;
	const Point p0(0.,0.,0.);
	Vector pCluster(p0);
	int nLayers = calorimeterGeo::numLayers() * calorimeterGeo::numViews();
	
	std::vector<double> eneLayer(nLayers,0.);
	std::vector<Vector> pLayer(nLayers);
		
	
	
	for (int jlog = 0; jlog < nLogs ; jlog++) {
		CalRecLog* recLog = m_CalRecLogs->Log(jlog);

		double eneLog = recLog->energy();
		Vector pLog = recLog->position() - p0;
		int layer = nLayers-1 - (recLog->layer() * 2 + recLog->view());
		eneLayer[layer]+=eneLog;
		pLayer[layer] += eneLog*pLog;
		ene  += eneLog;
		pCluster += eneLog*pLog;				
	}

	pCluster *= (1./ene);
	int i = 0;
	for( ;i<nLayers;i++){
			if(eneLayer[i]>0)pLayer[i] *= (1./eneLayer[i]);
			else pLayer[i]=p0;
	}


	CsICluster* cl = new CsICluster(ene,pCluster+p0);
	m_CsIClusterList->add(cl);
	cl->setEneLayer(eneLayer);
	cl->setPosLayer(pLayer);

	// defines global variable to be used for fcn
	g_elayer.resize(nLayers);
	for ( i =0;i<nLayers;i++)
	{
		// We are working in GeV
		g_elayer[i] = eneLayer[i]/1000.;
	}
	nbins = nLayers;
	slope = 1;  // temporary

	// Do profile fitting
    Profile(ene,cl);
	
}
//################################################
void CalClustersAlg::finalize()
//################################################
{
	m_CsIClusterList->writeOut();
}
