
#include "Event/dataManager.h"
#include "Event/messageManager.h"
#include "instrument/CalBase.h"
#include "instrument/calorimeterDetGeo.h"
#include "instrument/CalPedCalib.h"
#include "instrument/CalCalibLogs.h"
#include "reconstruction/CalRecLogsAlg.h"
#include "instrument/CalADCLogs.h"
#include "reconstruction/CalRecLogs.h"



//################################################
void CalRecLogsAlg::initialize()
//################################################
{
	
	m_CalPedLogs   = dataManager::instance()->getData("CalPedCalib",m_CalPedLogs);
	m_CalCalibLogs = dataManager::instance()->getData("CalCalibLogs",m_CalCalibLogs);
	m_CalGeoLogs   = dataManager::instance()->geo()->cal();

	m_CalRawLogs   = dataManager::instance()->getData("CalRawLogs",m_CalRawLogs);
	m_CalRecLogs   = dataManager::instance()->getData("CalRecLogs",m_CalRecLogs);
	m_CalRecLogs->clear();
}
//################################################
void CalRecLogsAlg::execute()
//################################################
{
	int nLogs = m_CalRawLogs->num();
	for (int jlog = 0; jlog < nLogs ; jlog++) {
		
		CalADCLog* ADCLog = m_CalRawLogs->Log(jlog);

		int ilayer          = ADCLog->layer();
		detGeo::axis view   = ADCLog->view();
	    int icol            = ADCLog->column();

		detGeo*    geoLog = m_CalGeoLogs->getLog(ilayer,view,icol);
		CalADCLog* pedLog = m_CalPedLogs->getLogID(CalLogID::ID(ilayer,view,icol));
		CalCalibLog* calibLog = m_CalCalibLogs->getLogID(CalLogID::ID(ilayer,view,icol));

		CalRecLog* recLog = m_CalRecLogs->getLogID(CalLogID::ID(ilayer,view,icol));
		computeEnergy(recLog, ADCLog, pedLog, calibLog);
		computePosition(recLog,geoLog, calibLog);
	}
}
//################################################
void CalRecLogsAlg::finalize()
//################################################
{
	m_CalRecLogs->writeOut();
}
//----------------- private ----------------------
//################################################
void CalRecLogsAlg::computeEnergy(CalRecLog* recLog, const CalADCLog* ADCLog,
								  const CalADCLog* pedLog, const CalCalibLog* calibLog)
//################################################
{
	
	double MAXPED = 4095.;
	double ADCSATURATION = 3850;
	bool eneSet = false;
	for (int irange = 0; irange < CALNRANGES ; irange++) {
		CalBase::RANGE r = (irange == 0? CalBase::LEX : CalBase::LE);
		double eneNeg = 0.;
    	double enePos = 0.;
		double adcNeg = 0.;
		double adcPos = 0.;
		double adcSatNeg = 0.;
		double adcSatPos = 0.;
		for (int iside = 0; iside < CALNSIDES; iside++) {
			CalBase::SIDE s = (iside == 0? CalBase::NEG : CalBase::POS);
			if (irange > 1) r = (irange == 2? CalBase::HEX : CalBase::HE);
			double adc = ADCLog->ADC(s,r);
			double ped = pedLog->ADC(s,r);
			double adc_ped = adc - ped;
			//if (s== CalBase::NEG) recLog->setNegADC(r,(adc-ped)/(MAXPED-ped));
			//else recLog->setPosADC(r,(adc-ped)/(MAXPED-ped));
			if (s== CalBase::NEG) recLog->setNegADC(r,adc_ped);
			else recLog->setPosADC(r,adc_ped);

			double adcSat = 0.6*(calibLog->getRail(iside,irange));
			double ene = calibLog->adc_to_MeV(adc_ped,iside,irange);
			iside == 0? eneNeg = ene: enePos = ene;
			iside == 0? adcNeg = adc_ped: adcPos = adc_ped;
			iside == 0? adcSatNeg = adcSat: adcSatPos = adcSat;
		}
		recLog->setNegEnergy(r,eneNeg);
		recLog->setPosEnergy(r,enePos);
		if(adcNeg<50 || adcPos<50)eneNeg=enePos=0;

		if (!eneSet && (adcNeg < adcSatNeg && adcPos < adcSatPos)) {
			eneSet = true;
			recLog->setNegEnergy(eneNeg);
			recLog->setPosEnergy(enePos);
            recLog->setBestRange(r);
		}
	}
}
//################################################
void CalRecLogsAlg::computePosition(CalRecLog* recLog, const detGeo* geoLog,
									const CalCalibLog* calibLog)
//################################################
{
	Point pCenter = geoLog->position();
	Point pSize   = geoLog->size();
	
	detGeo::axis view = recLog->view();
	double xdir = 0.;
	double ydir = 0.;
	ydir = (view == detGeo::X? 1. : 0);
	xdir = (view == detGeo::Y? 1. : 0);

	Vector dirLog(xdir,ydir,0.); 
	
	
	double eneNeg = recLog->negEnergy();
	double enePos = recLog->posEnergy();
	double asym = recLog->asymmetry();
	double slope = calibLog->getSlope(recLog->bestRange());

	Point pLog = pCenter+dirLog*asym*slope;

	recLog->setPosition(pLog);
}
