//-------------------------------------------------------------------
//
//     SiRecObjsAlg:
//
//	    Steers the Silicon-Tracker Reconstruction	
//
//		      Bill Atwood
//		      B. Atwood, JA Hernando, Santa Cruz, 02/05/99
//
//-------------------------------------------------------------------

#include <vector>
#include "reconstruction/SiRecObjsAlg.h"
#include "Event/dataManager.h"
#include "Event/messageManager.h"
#include "reconstruction/SiClusters.h"
#include "reconstruction/SiRecObjs.h"
#include "reconstruction/GFcandidates.h"
#include "reconstruction/CsIClusters.h"

//###################################################
void SiRecObjsAlg::initialize()
//###################################################
{
	if (!defined()) define();
	double MINENE = 0.03;

	m_SiRecObjs->clear();

	// For the moment:
	CsIClusterList* cal = 0;
	cal = dataManager::instance()->getData("CsIClusterList",cal);
	
	m_CsIEnergy = MINENE;
	m_CsIPosition = Point(0.,0.,0.);
	if (cal->num()) {
		double ene = 0.001*cal->Cluster(0)->energyCorrected();
		if (ene >= m_CsIEnergy) {
			m_CsIEnergy = ene;
			m_CsIPosition = cal->Cluster(0)->position();
		}
	} 
}

//###################################################
void SiRecObjsAlg::define()
//###################################################
{
	setDefine(true);
	m_SiClusters = dataManager::instance()->getData("SiClusters",m_SiClusters);
	m_SiRecObjs  = dataManager::instance()->getData("SiRecObjs",m_SiRecObjs);
}
//###################################################
void SiRecObjsAlg::execute()
//###################################################
{
	double MAXCLUSTERS = 250;
	GFtutor::load(m_SiClusters);
	if (m_SiClusters->nHits() >= MAXCLUSTERS) {
		messageManager::instance()->message(" not reconstructing tracks, too many clusters ");
		return;
	}

	// Example of how to flag = not use hits in Planes :
	// m_siClusters->flagHitsInPlane(SiCluster::X,6);
	// m_siClusters->writeOut();

	// search for gammas 
	//        The control parameters of the reconstruction are in:
	//            GFcontrol and GFtutor
	searchGammas();

	// search for tracks
	searchParticles();

}
//##############################################
void SiRecObjsAlg::finalize()
//##############################################
{
	m_SiRecObjs->writeOut();
}
//----------- private --------------------------

//##############################################
void SiRecObjsAlg::searchGammas()
//##############################################
{
    // Gamma reconstruction
    int ntries = 0;
	double factor = 1.;
	bool end = false;
    
	while (ntries < GFcontrol::gammaTries && !end) {

		double ene = m_CsIEnergy*factor;
		if (ene <= GFcontrol::minEnergy) ene = GFcontrol::minEnergy;

		GFcandidates* gamma = 
			new GFcandidates(GFcandidates::GAMMA, ene ,m_CsIPosition);

		if (gamma->numCandidates() > 0) {
			ntries++;

			int ican = 0;
			GFdata gammaCandidate = gamma->m_candidates[ican];

			int iniLayer = gammaCandidate.firstLayer();
			Ray testRay = gammaCandidate.ray();

			GFgamma* _GFgamma = new GFgamma(GFcontrol::FEne, GFcontrol::sigmaCut,
				ene, iniLayer, testRay);

			if (!_GFgamma->empty()) {
				_GFgamma->flagAllHits();
				m_SiRecObjs->addGamma(_GFgamma);
			} else delete _GFgamma;

		} else end = true;
    }

}

//###########################################################
void SiRecObjsAlg::searchParticles()
//###########################################################
{
    // Other tracks reconstruction
    int ntries = 0;
	double factor = 1./GFcontrol::particleTries;
    bool end = false;

    while (ntries < GFcontrol::particleTries && !end) {
        
		GFtutor::setVeto(false);
        
 		double ene = GFcontrol::FEneParticle*m_CsIEnergy*factor;
		if (ene <= GFcontrol::minEnergy) ene = GFcontrol::minEnergy;
		GFcandidates* tracks = 
			new GFcandidates(GFcandidates::PARTICLE, ene, m_CsIPosition);

		if (tracks->numCandidates() > 0) {
			ntries++;
			GFdata trackCandidate = tracks->m_candidates[0];

			int iniLayer = trackCandidate.firstLayer();
			Ray testRay = trackCandidate.ray();

			GFparticle* _GFparticle = 
				new GFparticle(GFcontrol::sigmaCut, ene, iniLayer, testRay);

			if (!_GFparticle->empty()) {
				m_SiRecObjs->addParticle(_GFparticle);		
				_GFparticle->flagAllHits();
			} else delete _GFparticle;

		} else end = true;

		GFtutor::setVeto(true);
    }
}
